#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import logging
import os

logger = logging.getLogger(__file__)


def setup_logging(verbosity: int):
    base_loglevel = int(os.getenv("LOGLEVEL", 30))
    verbosity = min(verbosity, 2)
    loglevel = base_loglevel - (verbosity * 10)
    logging.basicConfig(format="%(message)s")
    logging.getLogger().setLevel(logging.WARNING)
    logger.setLevel(loglevel)



import pytest
from ops import RelationNotFoundError
from ops.charm import CharmBase
from ops.framework import Framework

from scenario.state import Event, Network, Relation, State, _CharmSpec
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        _call = None
        called = False

        def __init__(self, framework: Framework):
            super().__init__(framework)

            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if MyCharm._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_ip_get(mycharm):
    mycharm._call = lambda *_: True

    def fetch_unit_address(charm: CharmBase):
        rel = charm.model.get_relation("metrics-endpoint")
        assert str(charm.model.get_binding(rel).network.bind_address) == "1.1.1.1"

    trigger(
        State(
            relations=[
                Relation(
                    interface="foo",
                    remote_app_name="remote",
                    endpoint="metrics-endpoint",
                    relation_id=1,
                ),
            ],
            networks=[Network.default("metrics-endpoint")],
        ),
        "update_status",
        mycharm,
        meta={
            "name": "foo",
            "requires": {"metrics-endpoint": {"interface": "foo"}},
        },
        post_event=fetch_unit_address,
    )


def test_no_relation_error(mycharm):
    """Attempting to call get_binding on a non-existing relation -> RelationNotFoundError"""
    mycharm._call = lambda *_: True

    def fetch_unit_address(charm: CharmBase):
        with pytest.raises(RelationNotFoundError):
            _ = charm.model.get_binding("foo").network

    trigger(
        State(
            relations=[
                Relation(
                    interface="foo",
                    remote_app_name="remote",
                    endpoint="metrics-endpoint",
                    relation_id=1,
                ),
            ],
            networks=[Network.default("metrics-endpoint")],
        ),
        "update_status",
        mycharm,
        meta={
            "name": "foo",
            "requires": {"metrics-endpoint": {"interface": "foo"}},
        },
        post_event=fetch_unit_address,
    )



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import MagicMock

import pytest
import yaml
from ops.charm import CharmBase, CharmEvents
from ops.framework import EventBase

from scenario import Context
from scenario.runtime import Runtime, UncaughtCharmError
from scenario.state import Event, Relation, State, _CharmSpec


def charm_type():
    class _CharmEvents(CharmEvents):
        pass

    class MyCharm(CharmBase):
        on = _CharmEvents()
        _event = None

        def __init__(self, framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._catchall)

        def _catchall(self, e):
            if self._event:
                return
            MyCharm._event = e

    return MyCharm


def test_event_emission():
    with TemporaryDirectory() as tempdir:
        meta = {
            "name": "foo",
            "requires": {"ingress-per-unit": {"interface": "ingress_per_unit"}},
        }

        my_charm_type = charm_type()

        class MyEvt(EventBase):
            pass

        my_charm_type.on.define_event("bar", MyEvt)

        runtime = Runtime(
            _CharmSpec(
                my_charm_type,
                meta=meta,
            ),
        )

        with runtime.exec(
            state=State(), event=Event("bar"), context=Context(my_charm_type, meta=meta)
        ) as ops:
            pass

        assert my_charm_type._event
        assert isinstance(my_charm_type._event, MyEvt)


@pytest.mark.parametrize("app_name", ("foo", "bar-baz", "QuX2"))
@pytest.mark.parametrize("unit_id", (1, 2, 42))
def test_unit_name(app_name, unit_id):
    meta = {
        "name": app_name,
    }

    my_charm_type = charm_type()

    runtime = Runtime(
        _CharmSpec(
            my_charm_type,
            meta=meta,
        ),
    )

    with runtime.exec(
        state=State(unit_id=unit_id),
        event=Event("start"),
        context=Context(my_charm_type, meta=meta),
    ) as ops:
        assert ops.charm.unit.name == f"{app_name}/{unit_id}"


def test_env_cleanup_on_charm_error():
    meta = {"name": "frank", "requires": {"box": {"interface": "triangle"}}}

    my_charm_type = charm_type()

    runtime = Runtime(
        _CharmSpec(
            my_charm_type,
            meta=meta,
        ),
    )

    with pytest.raises(UncaughtCharmError):
        with runtime.exec(
            state=State(),
            event=Event("box_relation_changed", relation=Relation("box")),
            context=Context(my_charm_type, meta=meta),
        ):
            assert os.getenv("JUJU_REMOTE_APP")
            _ = 1 / 0  # raise some error

    assert os.getenv("JUJU_REMOTE_APP", None) is None



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
import typing
from itertools import chain
from typing import Any, Callable, Dict, Iterable, Optional, TextIO, Type, Union

from scenario import Context
from scenario.logger import logger as scenario_logger
from scenario.state import (
    ATTACH_ALL_STORAGES,
    BREAK_ALL_RELATIONS,
    CREATE_ALL_RELATIONS,
    DETACH_ALL_STORAGES,
    META_EVENTS,
    Event,
    InjectRelation,
    State,
)

if typing.TYPE_CHECKING:
    from ops.testing import CharmType

CharmMeta = Optional[Union[str, TextIO, dict]]

logger = scenario_logger.getChild("scenario")


def decompose_meta_event(meta_event: Event, state: State):
    # decompose the meta event

    if meta_event.name in [ATTACH_ALL_STORAGES, DETACH_ALL_STORAGES]:
        logger.warning(f"meta-event {meta_event.name} not supported yet")
        return

    if meta_event.name in [CREATE_ALL_RELATIONS, BREAK_ALL_RELATIONS]:
        for relation in state.relations:
            event = Event(
                relation.endpoint + META_EVENTS[meta_event.name],
                args=(
                    # right now, the Relation object hasn't been created by ops yet, so we
                    # can't pass it down.
                    # this will be replaced by a Relation instance before the event is fired.
                    InjectRelation(relation.endpoint, relation.relation_id),
                ),
            )
            logger.debug(f"decomposed meta {meta_event.name}: {event}")
            yield event, state.copy()

    else:
        raise RuntimeError(f"unknown meta-event {meta_event.name}")


def generate_startup_sequence(state_template: State):
    yield from chain(
        decompose_meta_event(Event(ATTACH_ALL_STORAGES), state_template.copy()),
        ((Event("start"), state_template.copy()),),
        decompose_meta_event(Event(CREATE_ALL_RELATIONS), state_template.copy()),
        (
            (
                Event(
                    "leader_elected"
                    if state_template.leader
                    else "leader_settings_changed",
                ),
                state_template.copy(),
            ),
            (Event("config_changed"), state_template.copy()),
            (Event("install"), state_template.copy()),
        ),
    )


def generate_teardown_sequence(state_template: State):
    yield from chain(
        decompose_meta_event(Event(BREAK_ALL_RELATIONS), state_template.copy()),
        decompose_meta_event(Event(DETACH_ALL_STORAGES), state_template.copy()),
        (
            (Event("stop"), state_template.copy()),
            (Event("remove"), state_template.copy()),
        ),
    )


def generate_builtin_sequences(template_states: Iterable[State]):
    for template_state in template_states:
        yield from chain(
            generate_startup_sequence(template_state),
            generate_teardown_sequence(template_state),
        )


def check_builtin_sequences(
    charm_type: Type["CharmType"],
    meta: Optional[Dict[str, Any]] = None,
    actions: Optional[Dict[str, Any]] = None,
    config: Optional[Dict[str, Any]] = None,
    template_state: State = None,
    pre_event: Optional[Callable[["CharmType"], None]] = None,
    post_event: Optional[Callable[["CharmType"], None]] = None,
) -> object:
    """Test that all the builtin startup and teardown events can fire without errors.

    This will play both scenarios with and without leadership, and raise any exceptions.

    This is a baseline check that in principle all charms (except specific use-cases perhaps),
    should pass out of the box.

    If you want to, you can inject more stringent state checks using the
    pre_event and post_event hooks.
    """

    template = template_state if template_state else State()
    out = []

    for event, state in generate_builtin_sequences(
        (
            template.replace(leader=True),
            template.replace(leader=False),
        ),
    ):
        ctx = Context(charm_type=charm_type, meta=meta, actions=actions, config=config)
        out.append(
            ctx.run(
                event,
                state=state,
                pre_event=pre_event,
                post_event=post_event,
            ),
        )
    return out



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
from pathlib import Path

from scenario.scripts.errors import InvalidTargetUnitName


class JujuUnitName(str):
    """This class represents the name of a juju unit that can be snapshotted."""

    def __init__(self, unit_name: str):
        super().__init__()
        app_name, _, unit_id = unit_name.rpartition("/")
        if not app_name or not unit_id:
            raise InvalidTargetUnitName(f"invalid unit name: {unit_name!r}")
        self.unit_name = unit_name
        self.app_name = app_name
        self.unit_id = int(unit_id)
        self.normalized = f"{app_name}-{unit_id}"
        self.remote_charm_root = Path(
            f"/var/lib/juju/agents/unit-{self.normalized}/charm",
        )



import pytest
from ops.charm import ActionEvent, CharmBase
from ops.framework import Framework

from scenario import Context
from scenario.context import InvalidEventError
from scenario.state import Action, Event, State
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        _evt_handler = None

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if handler := self._evt_handler:
                handler(event)

    return MyCharm


@pytest.mark.parametrize("baz_value", (True, False))
def test_action_event(mycharm, baz_value):
    ctx = Context(
        mycharm,
        meta={"name": "foo"},
        actions={
            "foo": {"params": {"bar": {"type": "number"}, "baz": {"type": "boolean"}}}
        },
    )
    action = Action("foo", params={"baz": baz_value, "bar": 10})
    ctx.run_action(action, State())

    evt = ctx.emitted_events[0]

    assert evt.params["bar"] == 10
    assert evt.params["baz"] is baz_value


def test_action_pre_post(mycharm):
    ctx = Context(
        mycharm,
        meta={"name": "foo"},
        actions={
            "foo": {"params": {"bar": {"type": "number"}, "baz": {"type": "boolean"}}}
        },
    )
    action = Action("foo", params={"baz": True, "bar": 10})
    ctx.run_action(
        action,
        State(),
        pre_event=lambda charm: None,
        post_event=lambda charm: None,
    )


@pytest.mark.parametrize("res_value", ("one", 1, [2], ["bar"], (1,), {1, 2}))
def test_action_event_results_invalid(mycharm, res_value):
    def handle_evt(charm: CharmBase, evt: ActionEvent):
        with pytest.raises((TypeError, AttributeError)):
            evt.set_results(res_value)

    mycharm._evt_handler = handle_evt

    action = Action("foo")
    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    ctx.run_action(action, State())


def test_cannot_run_action(mycharm):
    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    action = Action("foo")

    with pytest.raises(InvalidEventError):
        ctx.run(action, state=State())


def test_cannot_run_action_name(mycharm):
    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    action = Action("foo")
    with pytest.raises(InvalidEventError):
        ctx.run(action.event.name, state=State())


def test_cannot_run_action_event(mycharm):
    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    action = Action("foo")
    with pytest.raises(InvalidEventError):
        ctx.run(action.event, state=State())


@pytest.mark.parametrize("res_value", ({"a": {"b": {"c"}}}, {"d": "e"}))
def test_action_event_results_valid(mycharm, res_value):
    def handle_evt(charm: CharmBase, evt):
        if not isinstance(evt, ActionEvent):
            return
        evt.set_results(res_value)
        evt.log("foo")
        evt.log("bar")

    mycharm._evt_handler = handle_evt

    action = Action("foo")
    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})

    out = ctx.run_action(action, State())

    assert out.results == res_value
    assert out.success is True


@pytest.mark.parametrize("res_value", ({"a": {"b": {"c"}}}, {"d": "e"}))
def test_action_event_outputs(mycharm, res_value):
    def handle_evt(charm: CharmBase, evt: ActionEvent):
        if not isinstance(evt, ActionEvent):
            return

        evt.set_results({"my-res": res_value})
        evt.log("log1")
        evt.log("log2")
        evt.fail("failed becozz")

    mycharm._evt_handler = handle_evt

    action = Action("foo")
    ctx = Context(mycharm, meta={"name": "foo"}, actions={"foo": {}})
    out = ctx.run_action(action, State())

    assert out.failure == "failed becozz"
    assert out.logs == ["log1", "log2"]
    assert out.success is False



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import datetime
import json
import os
import re
import shlex
import sys
import tempfile
from dataclasses import asdict, dataclass
from enum import Enum
from itertools import chain
from pathlib import Path
from subprocess import run
from typing import Any, BinaryIO, Dict, Iterable, List, Optional, TextIO, Tuple, Union

import ops.pebble
import typer
import yaml
from ops.storage import SQLiteStorage

from scenario.runtime import UnitStateDB
from scenario.scripts.errors import InvalidTargetModelName, InvalidTargetUnitName
from scenario.scripts.logger import logger as root_scripts_logger
from scenario.scripts.utils import JujuUnitName
from scenario.state import (
    Address,
    BindAddress,
    BindFailedError,
    Container,
    Event,
    Model,
    Mount,
    Network,
    Port,
    Relation,
    Secret,
    State,
    _EntityStatus,
)

logger = root_scripts_logger.getChild(__file__)

JUJU_RELATION_KEYS = frozenset({"egress-subnets", "ingress-address", "private-address"})
JUJU_CONFIG_KEYS = frozenset({})

SNAPSHOT_OUTPUT_DIR = (Path(os.getcwd()).parent / "snapshot_storage").absolute()
CHARM_SUBCLASS_REGEX = re.compile(r"class (\D+)\(CharmBase\):")


def _try_format(string: str):
    try:
        import black

        try:
            return black.format_str(string, mode=black.Mode())
        except black.parsing.InvalidInput as e:
            logger.error(f"error parsing {string}: {e}")
            return string
    except ModuleNotFoundError:
        logger.warning("install black for formatting")
        return string


def format_state(state: State):
    """Stringify this State as nicely as possible."""
    return _try_format(repr(state))


PYTEST_TEST_TEMPLATE = """
from scenario import *
from charm import {ct}

def test_case():
    # Arrange: prepare the state
    state = {state}

    #Act: trigger an event on the state
    ctx = Context(
        {ct},
        juju_version="{jv}")

    out = ctx.run(
        {en}
        state,
        )

    # Assert: verify that the output state is the way you want it to be
    # TODO: add assertions
"""


def format_test_case(
    state: State,
    charm_type_name: str = None,
    event_name: str = None,
    juju_version: str = None,
):
    """Format this State as a pytest test case."""
    ct = charm_type_name or "CHARM_TYPE,  # TODO: replace with charm type name"
    en = "EVENT_NAME,  # TODO: replace with event name"
    if event_name:
        try:
            en = Event(event_name).bind(state)
        except BindFailedError:
            logger.error(
                f"Failed to bind {event_name} to {state}; leaving placeholder instead",
            )

    jv = juju_version or "3.0,  # TODO: check juju version is correct"
    state_fmt = repr(state)
    return _try_format(
        PYTEST_TEST_TEMPLATE.format(state=state_fmt, ct=ct, en=en, jv=jv),
    )


def _juju_run(cmd: str, model=None) -> Dict[str, Any]:
    """Execute juju {command} in a given model."""
    _model = f" -m {model}" if model else ""
    cmd = f"juju {cmd}{_model} --format json"
    raw = run(shlex.split(cmd), capture_output=True, text=True).stdout
    return json.loads(raw)


def _juju_ssh(target: JujuUnitName, cmd: str, model: Optional[str] = None) -> str:
    _model = f" -m {model}" if model else ""
    command = f"juju ssh{_model} {target.unit_name} {cmd}"
    raw = run(shlex.split(command), capture_output=True, text=True).stdout
    return raw


def _juju_exec(target: JujuUnitName, model: Optional[str], cmd: str) -> str:
    """Execute a juju command.

    Notes:
        Visit the Juju documentation to view all possible Juju commands:
        https://juju.is/docs/olm/juju-cli-commands
    """
    _model = f" -m {model}" if model else ""
    _target = f" -u {target}" if target else ""
    return run(
        shlex.split(f"juju exec{_model}{_target} -- {cmd}"),
        capture_output=True,
        text=True,
    ).stdout


def get_leader(target: JujuUnitName, model: Optional[str]):
    # could also get it from _juju_run('status')...
    logger.info("getting leader...")
    return _juju_exec(target, model, "is-leader") == "True"


def get_network(target: JujuUnitName, model: Optional[str], endpoint: str) -> Network:
    """Get the Network data structure for this endpoint."""
    raw = _juju_exec(target, model, f"network-get {endpoint}")
    json_data = yaml.safe_load(raw)

    bind_addresses = []
    for raw_bind in json_data["bind-addresses"]:
        addresses = []
        for raw_adds in raw_bind["addresses"]:
            addresses.append(
                Address(
                    hostname=raw_adds["hostname"],
                    value=raw_adds["value"],
                    cidr=raw_adds["cidr"],
                    address=raw_adds.get("address", ""),
                ),
            )

        bind_addresses.append(
            BindAddress(
                interface_name=raw_bind.get("interface-name", ""),
                addresses=addresses,
            ),
        )
    return Network(
        name=endpoint,
        bind_addresses=bind_addresses,
        egress_subnets=json_data.get("egress-subnets", None),
        ingress_addresses=json_data.get("ingress-addresses", None),
    )


def get_secrets(
    target: JujuUnitName,  # noqa: U100
    model: Optional[str],  # noqa: U100
    metadata: Dict,  # noqa: U100
    relations: Tuple[str, ...] = (),  # noqa: U100
) -> List[Secret]:
    """Get Secret list from the charm."""
    logger.warning("Secrets snapshotting not implemented yet. Also, are you *sure*?")
    return []


def get_networks(
    target: JujuUnitName,
    model: Optional[str],
    metadata: Dict,
    include_dead: bool = False,
    relations: Tuple[str, ...] = (),
) -> List[Network]:
    """Get all Networks from this unit."""
    logger.info("getting networks...")
    networks = [get_network(target, model, "juju-info")]

    endpoints = relations  # only alive relations
    if include_dead:
        endpoints = chain(
            metadata.get("provides", ()),
            metadata.get("requires", ()),
            metadata.get("peers", ()),
        )

    for endpoint in endpoints:
        logger.debug(f"  getting network for endpoint {endpoint!r}")
        networks.append(get_network(target, model, endpoint))
    return networks


def get_metadata(target: JujuUnitName, model: Model):
    """Get metadata.yaml from this target."""
    logger.info("fetching metadata...")

    meta_path = target.remote_charm_root / "metadata.yaml"

    raw_meta = _juju_ssh(
        target,
        f"cat {meta_path}",
        model=model.name,
    )
    return yaml.safe_load(raw_meta)


class RemotePebbleClient:
    """Clever little class that wraps calls to a remote pebble client."""

    def __init__(
        self,
        container: str,
        target: JujuUnitName,
        model: Optional[str] = None,
    ):
        self.socket_path = f"/charm/containers/{container}/pebble.socket"
        self.container = container
        self.target = target
        self.model = model

    def _run(self, cmd: str) -> str:
        _model = f" -m {self.model}" if self.model else ""
        command = (
            f"juju ssh{_model} --container {self.container} {self.target.unit_name} "
            f"/charm/bin/pebble {cmd}"
        )
        proc = run(shlex.split(command), capture_output=True, text=True)
        if proc.returncode == 0:
            return proc.stdout
        raise RuntimeError(
            f"error wrapping pebble call with {command}: "
            f"process exited with {proc.returncode}; "
            f"stdout = {proc.stdout}; "
            f"stderr = {proc.stderr}",
        )

    def can_connect(self) -> bool:
        try:
            version = self.get_system_info()
        except Exception:
            return False
        return bool(version)

    def get_system_info(self):
        return self._run("version")

    def get_plan(self) -> dict:
        plan_raw = self._run("plan")
        return yaml.safe_load(plan_raw)

    def pull(
        self,
        path: str,  # noqa: U100
        *,
        encoding: Optional[str] = "utf-8",  # noqa: U100
    ) -> Union[BinaryIO, TextIO]:
        raise NotImplementedError()

    def list_files(
        self,
        path: str,  # noqa: U100
        *,
        pattern: Optional[str] = None,  # noqa: U100
        itself: bool = False,  # noqa: U100
    ) -> List[ops.pebble.FileInfo]:
        raise NotImplementedError()

    def get_checks(
        self,
        level: Optional[ops.pebble.CheckLevel] = None,
        names: Optional[Iterable[str]] = None,
    ) -> List[ops.pebble.CheckInfo]:
        _level = f" --level={level}" if level else ""
        _names = (" " + " ".join(names)) if names else ""
        out = self._run(f"checks{_level}{_names}")
        if out == "Plan has no health checks.":
            return []
        raise NotImplementedError()


def fetch_file(
    target: JujuUnitName,
    remote_path: Union[Path, str],
    container_name: str,
    local_path: Union[Path, str],
    model: Optional[str] = None,
) -> None:
    """Download a file from a live unit to a local path."""
    model_arg = f" -m {model}" if model else ""
    scp_cmd = (
        f"juju scp --container {container_name}{model_arg} "
        f"{target.unit_name}:{remote_path} {local_path}"
    )
    run(shlex.split(scp_cmd))


def get_mounts(
    target: JujuUnitName,
    model: Optional[str],
    container_name: str,
    container_meta: Dict,
    fetch_files: Optional[List[Path]] = None,
    temp_dir_base_path: Path = SNAPSHOT_OUTPUT_DIR,
) -> Dict[str, Mount]:
    """Get named Mounts from a container's metadata, and download specified files from the unit."""
    mount_meta = container_meta.get("mounts")

    if fetch_files and not mount_meta:
        logger.error(
            f"No mounts defined for container {container_name} in metadata.yaml. "
            f"Cannot fetch files {fetch_files} for this container.",
        )
        return {}

    mount_spec = {}
    for mt in mount_meta or ():
        if name := mt.get("storage"):
            mount_spec[name] = mt["location"]
        else:
            logger.error(f"unknown mount type: {mt}")

    mounts = {}
    for remote_path in fetch_files or ():
        found = None
        for mn, mt in mount_spec.items():
            if str(remote_path).startswith(mt):
                found = mn, mt

        if not found:
            logger.error(
                "could not find mount corresponding to requested remote_path "
                f"{remote_path}: skipping...",
            )
            continue

        mount_name, src = found
        mount = mounts.get(mount_name)
        if not mount:
            # create the mount obj and tempdir
            location = tempfile.TemporaryDirectory(prefix=str(temp_dir_base_path)).name
            mount = Mount(src=src, location=location)
            mounts[mount_name] = mount

        # populate the local tempdir
        filepath = Path(mount.location).joinpath(*remote_path.parts[1:])
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        try:
            fetch_file(
                target,
                container_name=container_name,
                model=model,
                remote_path=remote_path,
                local_path=filepath,
            )

        except RuntimeError as e:
            logger.error(e)

    return mounts


def get_container(
    target: JujuUnitName,
    model: Optional[str],
    container_name: str,
    container_meta: Dict,
    fetch_files: Optional[List[Path]] = None,
    temp_dir_base_path: Path = SNAPSHOT_OUTPUT_DIR,
) -> Container:
    """Get container data structure from the target."""
    remote_client = RemotePebbleClient(container_name, target, model)
    plan = remote_client.get_plan()

    return Container(
        name=container_name,
        _base_plan=plan,
        can_connect=remote_client.can_connect(),
        mounts=get_mounts(
            target,
            model,
            container_name,
            container_meta,
            fetch_files,
            temp_dir_base_path=temp_dir_base_path,
        ),
    )


def get_containers(
    target: JujuUnitName,
    model: Optional[str],
    metadata: Optional[Dict],
    fetch_files: Dict[str, List[Path]] = None,
    temp_dir_base_path: Path = SNAPSHOT_OUTPUT_DIR,
) -> List[Container]:
    """Get all containers from this unit."""
    fetch_files = fetch_files or {}
    logger.info("getting containers...")

    if not metadata:
        logger.warning("no metadata: unable to get containers")
        return []

    containers = []
    for container_name, container_meta in metadata.get("containers", {}).items():
        container = get_container(
            target,
            model,
            container_name,
            container_meta,
            fetch_files=fetch_files.get(container_name),
            temp_dir_base_path=temp_dir_base_path,
        )
        containers.append(container)
    return containers


def get_juju_status(model: Optional[str]) -> Dict:
    """Return juju status as json."""
    logger.info("getting status...")
    return _juju_run("status --relations", model=model)


@dataclass
class Status:
    app: _EntityStatus
    unit: _EntityStatus
    workload_version: str


def get_status(juju_status: Dict, target: JujuUnitName) -> Status:
    """Parse `juju status` to get the Status data structure and some relation information."""
    app = juju_status["applications"][target.app_name]

    app_status_raw = app["application-status"]
    app_status = app_status_raw["current"], app_status_raw.get("message", "")

    unit_status_raw = app["units"][target]["workload-status"]
    unit_status = unit_status_raw["current"], unit_status_raw.get("message", "")

    workload_version = app.get("version", "")
    return Status(
        app=_EntityStatus(*app_status),
        unit=_EntityStatus(*unit_status),
        workload_version=workload_version,
    )


def get_endpoints(juju_status: Dict, target: JujuUnitName) -> Tuple[str, ...]:
    """Parse `juju status` to get the relation names owned by the target."""
    app = juju_status["applications"][target.app_name]
    relations_raw = app.get("relations", None)
    if not relations_raw:
        return ()
    relations = tuple(relations_raw.keys())
    return relations


def get_opened_ports(
    target: JujuUnitName,
    model: Optional[str],
) -> List[Port]:
    """Get opened ports list from target."""
    logger.info("getting opened ports...")

    opened_ports_raw = _juju_exec(
        target,
        model,
        "opened-ports --format json",
    )
    ports = []

    for raw_port in json.loads(opened_ports_raw):
        _port_n, _proto = raw_port.split("/")
        ports.append(Port(_proto, int(_port_n)))

    return ports


def get_config(
    target: JujuUnitName,
    model: Optional[str],
) -> Dict[str, Union[str, int, float, bool]]:
    """Get config dict from target."""

    logger.info("getting config...")
    json_data = _juju_run(f"config {target.app_name}", model=model)

    # dispatch table for builtin config options
    converters = {
        "string": str,
        "int": int,
        "integer": int,  # fixme: which one is it?
        "number": float,
        "boolean": lambda x: x == "true",
        "attrs": lambda x: x,  # fixme: wot?
    }

    cfg = {}
    for name, option in json_data.get("settings", ()).items():
        if value := option.get("value"):
            try:
                converter = converters[option["type"]]
            except KeyError:
                raise ValueError(f'unrecognized type {option["type"]}')
            cfg[name] = converter(value)

        else:
            logger.debug(f"skipped {name}: no value.")

    return cfg


def _get_interface_from_metadata(endpoint: str, metadata: Dict) -> Optional[str]:
    """Get the name of the interface used by endpoint."""
    for role in ["provides", "requires"]:
        for ep, ep_meta in metadata.get(role, {}).items():
            if ep == endpoint:
                return ep_meta["interface"]

    logger.error(f"No interface for endpoint {endpoint} found in charm metadata.")
    return None


def get_relations(
    target: JujuUnitName,
    model: Optional[str],
    metadata: Dict,
    include_juju_relation_data=False,
) -> List[Relation]:
    """Get the list of relations active for this target."""
    logger.info("getting relations...")

    try:
        json_data = _juju_run(f"show-unit {target}", model=model)
    except json.JSONDecodeError as e:
        raise InvalidTargetUnitName(target) from e

    def _clean(relation_data: dict):
        if include_juju_relation_data:
            return relation_data
        else:
            for key in JUJU_RELATION_KEYS:
                del relation_data[key]
        return relation_data

    relations = []
    for raw_relation in json_data[target].get("relation-info", ()):
        logger.debug(
            f"  getting relation data for endpoint {raw_relation.get('endpoint')!r}",
        )
        related_units = raw_relation.get("related-units")
        if not related_units:
            continue
        #    related-units:
        #      owner/0:
        #        in-scope: true
        #        data:
        #          egress-subnets: 10.152.183.130/32
        #          ingress-address: 10.152.183.130
        #          private-address: 10.152.183.130

        relation_id = raw_relation["relation-id"]

        local_unit_data_raw = _juju_exec(
            target,
            model,
            f"relation-get -r {relation_id} - {target} --format json",
        )
        local_unit_data = json.loads(local_unit_data_raw)
        local_app_data_raw = _juju_exec(
            target,
            model,
            f"relation-get -r {relation_id} - {target} --format json --app",
        )
        local_app_data = json.loads(local_app_data_raw)

        some_remote_unit_id = JujuUnitName(next(iter(related_units)))

        # fixme: at the moment the juju CLI offers no way to see what type of relation this is;
        #  if it's a peer relation or a subordinate, we should use the corresponding
        #  scenario.state types instead of a regular Relation.

        relations.append(
            Relation(
                endpoint=raw_relation["endpoint"],
                interface=_get_interface_from_metadata(
                    raw_relation["endpoint"],
                    metadata,
                ),
                relation_id=relation_id,
                remote_app_data=raw_relation["application-data"],
                remote_app_name=some_remote_unit_id.app_name,
                remote_units_data={
                    JujuUnitName(tgt).unit_id: _clean(val["data"])
                    for tgt, val in related_units.items()
                },
                local_app_data=local_app_data,
                local_unit_data=_clean(local_unit_data),
            ),
        )
    return relations


def get_model(name: str = None) -> Model:
    """Get the Model data structure."""
    logger.info("getting model...")

    json_data = _juju_run("models")
    model_name = name or json_data["current-model"]
    try:
        model_info = next(
            filter(lambda m: m["short-name"] == model_name, json_data["models"]),
        )
    except StopIteration as e:
        raise InvalidTargetModelName(name) from e

    model_uuid = model_info["model-uuid"]
    model_type = model_info["type"]

    return Model(name=model_name, uuid=model_uuid, type=model_type)


def try_guess_charm_type_name() -> Optional[str]:
    """If we are running this from a charm project root, get the charm type name from charm.py."""
    try:
        charm_path = Path(os.getcwd()) / "src" / "charm.py"
        if charm_path.exists():
            source = charm_path.read_text()
            charms = CHARM_SUBCLASS_REGEX.findall(source)
            if len(charms) < 1:
                raise RuntimeError(f"Not enough charms at {charm_path}.")
            elif len(charms) > 1:
                raise RuntimeError(f"Too many charms at {charm_path}.")
            return charms[0]
    except Exception as e:
        logger.warning(f"unable to guess charm type: {e}")
    return None


class FormatOption(
    str,
    Enum,
):  # Enum for typer support, str for native comparison and ==.
    """Output formatting options for snapshot."""

    state = "state"  # the default: will print the python repr of the State dataclass.
    json = "json"
    pytest = "pytest"


def get_juju_version(juju_status: Dict) -> str:
    """Get juju agent version from juju status output."""
    return juju_status["model"]["version"]


def get_charm_version(target: JujuUnitName, juju_status: Dict) -> str:
    """Get charm version info from juju status output."""
    app_info = juju_status["applications"][target.app_name]
    channel = app_info.get("charm-channel", "<local charm>")
    charm_name = app_info.get("charm-name", "n/a")
    workload_version = app_info.get("version", "n/a")
    charm_rev = app_info.get("charm-rev", "n/a")
    charm_origin = app_info.get("charm-origin", "n/a")
    return (
        f"charm {charm_name!r} ({channel}/{charm_rev}); "
        f"origin := {charm_origin}; app version := {workload_version}."
    )


class RemoteUnitStateDB(UnitStateDB):
    """Represents a remote unit's state db."""

    def __init__(self, model: Optional[str], target: JujuUnitName):
        self._tempfile = tempfile.NamedTemporaryFile()
        super().__init__(self._tempfile.name)

        self._model = model
        self._target = target

    def _fetch_state(self):
        fetch_file(
            self._target,
            remote_path=self._target.remote_charm_root / ".unit-state.db",
            container_name="charm",
            local_path=self._state_file,
            model=self._model,
        )

    @property
    def _has_state(self):
        """Whether the state file exists."""
        return self._state_file.exists() and self._state_file.read_bytes()

    def _open_db(self) -> SQLiteStorage:
        if not self._has_state:
            self._fetch_state()
        return super()._open_db()


def _snapshot(
    target: str,
    model: Optional[str] = None,
    pprint: bool = True,
    include: Optional[str] = None,
    include_juju_relation_data=False,
    include_dead_relation_networks=False,
    format: FormatOption = "state",
    event_name: Optional[str] = None,
    fetch_files: Optional[Dict[str, List[Path]]] = None,
    temp_dir_base_path: Path = SNAPSHOT_OUTPUT_DIR,
):
    """see snapshot's docstring"""

    try:
        target = JujuUnitName(target)
    except InvalidTargetUnitName:
        logger.critical(
            f"invalid target: {target!r} is not a valid unit name. Should be formatted like so:"
            f"`foo/1`, or `database/0`, or `myapp-foo-bar/42`.",
        )
        sys.exit(1)

    logger.info(f'beginning snapshot of {target} in model {model or "<current>"}...')

    def if_include(key, fn, default):
        if include is None or key in include:
            return fn()
        return default

    try:
        state_model = get_model(model)
    except InvalidTargetModelName:
        logger.critical(f"unable to get Model from name {model}.", exc_info=True)
        sys.exit(1)

    # todo: what about controller?
    model = state_model.name

    metadata = get_metadata(target, state_model)
    if not metadata:
        logger.critical(f"could not fetch metadata from {target}.")
        sys.exit(1)

    try:
        unit_state_db = RemoteUnitStateDB(model, target)
        juju_status = get_juju_status(model)
        endpoints = get_endpoints(juju_status, target)
        status = get_status(juju_status, target=target)

        state = State(
            leader=get_leader(target, model),
            unit_status=status.unit,
            app_status=status.app,
            workload_version=status.workload_version,
            model=state_model,
            config=if_include("c", lambda: get_config(target, model), {}),
            opened_ports=if_include(
                "p",
                lambda: get_opened_ports(target, model),
                [],
            ),
            relations=if_include(
                "r",
                lambda: get_relations(
                    target,
                    model,
                    metadata=metadata,
                    include_juju_relation_data=include_juju_relation_data,
                ),
                [],
            ),
            containers=if_include(
                "k",
                lambda: get_containers(
                    target,
                    model,
                    metadata,
                    fetch_files=fetch_files,
                    temp_dir_base_path=temp_dir_base_path,
                ),
                [],
            ),
            networks=if_include(
                "n",
                lambda: get_networks(
                    target,
                    model,
                    metadata,
                    include_dead=include_dead_relation_networks,
                    relations=endpoints,
                ),
                [],
            ),
            secrets=if_include(
                "S",
                lambda: get_secrets(
                    target,
                    model,
                    metadata,
                    relations=endpoints,
                ),
                [],
            ),
            deferred=if_include(
                "d",
                unit_state_db.get_deferred_events,
                [],
            ),
            stored_state=if_include(
                "t",
                unit_state_db.get_stored_state,
                [],
            ),
        )

        # todo: these errors should surface earlier.
    except InvalidTargetUnitName:
        _model = f"model {model}" or "the current model"
        logger.critical(f"invalid target: {target!r} not found in {_model}")
        sys.exit(1)
    except InvalidTargetModelName:
        logger.critical(f"invalid model: {model!r} not found.")
        sys.exit(1)

    logger.info("snapshot done.")

    if pprint:
        charm_version = get_charm_version(target, juju_status)
        juju_version = get_juju_version(juju_status)
        if format == FormatOption.pytest:
            charm_type_name = try_guess_charm_type_name()
            txt = format_test_case(
                state,
                event_name=event_name,
                charm_type_name=charm_type_name,
                juju_version=juju_version,
            )
        elif format == FormatOption.state:
            txt = format_state(state)
        elif format == FormatOption.json:
            txt = json.dumps(asdict(state), indent=2)
        else:
            raise ValueError(f"unknown format {format}")

        # json does not support comments, so it would be invalid output.
        if format != FormatOption.json:
            # print out some metadata
            controller_timestamp = juju_status["controller"]["timestamp"]
            local_timestamp = datetime.datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
            print(
                f"# Generated by scenario.snapshot. \n"
                f"# Snapshot of {state_model.name}:{target.unit_name} at {local_timestamp}. \n"
                f"# Controller timestamp := {controller_timestamp}. \n"
                f"# Juju version := {juju_version} \n"
                f"# Charm fingerprint := {charm_version} \n",
            )

        print(txt)

    return state


def snapshot(
    target: str = typer.Argument(..., help="Target unit."),
    model: Optional[str] = typer.Option(
        None,
        "-m",
        "--model",
        help="Which model to look at.",
    ),
    format: FormatOption = typer.Option(
        "state",
        "-f",
        "--format",
        help="How to format the output. "
        "``state``: Outputs a black-formatted repr() of the State object (if black is installed! "
        "else it will be ugly but valid python code). All you need to do then is import the "
        "necessary objects from scenario.state, and you should have a valid State object. "
        "``json``: Outputs a Jsonified State object. Perfect for storage. "
        "``pytest``: Outputs a full-blown pytest scenario test based on this State. "
        "Pipe it to a file and fill in the blanks.",
    ),
    event_name: str = typer.Option(
        None,
        "--event_name",
        "-e",
        help="Event to include in the generate test file; only applicable "
        "if the output format is 'pytest'.",
    ),
    include: str = typer.Option(
        "rckndtp",
        "--include",
        "-i",
        help="What data to include in the state. "
        "``r``: relation, ``c``: config, ``k``: containers, "
        "``n``: networks, ``S``: secrets(!), ``p``: opened ports, "
        "``d``: deferred events, ``t``: stored state.",
    ),
    include_dead_relation_networks: bool = typer.Option(
        False,
        "--include-dead-relation-networks",
        help="Whether to gather networks of inactive relation endpoints.",
        is_flag=True,
    ),
    include_juju_relation_data: bool = typer.Option(
        False,
        "--include-juju-relation-data",
        help="Whether to include in the relation data the default juju keys (egress-subnets,"
        "ingress-address, private-address).",
        is_flag=True,
    ),
    fetch: Path = typer.Option(
        None,
        "--fetch",
        help="Path to a local file containing a json spec of files to be fetched from the unit. "
        "For k8s units, it's supposed to be a {container_name: List[Path]} mapping listing "
        "the files that need to be fetched from the existing containers.",
    ),
    # TODO: generalize "fetch" to allow passing '.' for the 'charm' container or 'the machine'.
    output_dir: Path = typer.Option(
        SNAPSHOT_OUTPUT_DIR,
        "--output-dir",
        help="Directory in which to store any files fetched as part of the state. In the case "
        "of k8s charms, this might mean files obtained through Mounts,",
    ),
) -> State:
    """Gather and output the State of a remote target unit.

    If black is available, the output will be piped through it for formatting.

    Usage: snapshot myapp/0 > ./tests/scenario/case1.py
    """

    fetch_files = json.loads(fetch.read_text()) if fetch else None

    return _snapshot(
        target=target,
        model=model,
        format=format,
        event_name=event_name,
        include=include,
        include_juju_relation_data=include_juju_relation_data,
        include_dead_relation_networks=include_dead_relation_networks,
        temp_dir_base_path=output_dir,
        fetch_files=fetch_files,
    )


# for the benefit of script usage
_snapshot.__doc__ = snapshot.__doc__

if __name__ == "__main__":
    # print(_snapshot("zookeeper/0", model="foo", format=FormatOption.pytest))

    print(
        _snapshot(
            "traefik/0",
            format=FormatOption.state,
            include="r",
            # fetch_files={
            #     "traefik": [
            #         Path("/opt/traefik/juju/certificates.yaml"),
            #         Path("/opt/traefik/juju/certificate.cert"),
            #         Path("/opt/traefik/juju/certificate.key"),
            #         Path("/etc/traefik/traefik.yaml"),
            #     ]
            # },
        ),
    )



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import typing
from contextlib import contextmanager
from typing import ContextManager, List, Type, TypeVar

from ops import CollectStatusEvent
from ops.framework import (
    CommitEvent,
    EventBase,
    Framework,
    Handle,
    NoTypeError,
    PreCommitEvent,
)

_T = TypeVar("_T", bound=EventBase)


@contextmanager
def capture_events(
    *types: Type[EventBase],
    include_framework=False,
    include_deferred=True,
) -> ContextManager[List[EventBase]]:
    """Capture all events of type `*types` (using instance checks).

    Arguments exposed so that you can define your own fixtures if you want to.

    Example::
    >>> from ops.charm import StartEvent
    >>> from scenario import Event, State
    >>> from charm import MyCustomEvent, MyCharm  # noqa
    >>>
    >>> def test_my_event():
    >>>     with capture_events(StartEvent, MyCustomEvent) as captured:
    >>>         trigger(State(), ("start", MyCharm, meta=MyCharm.META)
    >>>
    >>>     assert len(captured) == 2
    >>>     e1, e2 = captured
    >>>     assert isinstance(e2, MyCustomEvent)
    >>>     assert e2.custom_attr == 'foo'
    """
    allowed_types = types or (EventBase,)

    captured = []
    _real_emit = Framework._emit
    _real_reemit = Framework.reemit

    def _wrapped_emit(self, evt):
        if not include_framework and isinstance(
            evt,
            (PreCommitEvent, CommitEvent, CollectStatusEvent),
        ):
            return _real_emit(self, evt)

        if isinstance(evt, allowed_types):
            # dump/undump the event to ensure any custom attributes are (re)set by restore()
            evt.restore(evt.snapshot())
            captured.append(evt)

        return _real_emit(self, evt)

    def _wrapped_reemit(self):
        # Framework calls reemit() before emitting the main juju event. We intercept that call
        # and capture all events in storage.

        if not include_deferred:
            return _real_reemit(self)

        # load all notices from storage as events.
        for event_path, _, _ in self._storage.notices():
            event_handle = Handle.from_path(event_path)
            try:
                event = self.load_snapshot(event_handle)
            except NoTypeError:
                continue
            event = typing.cast(EventBase, event)
            event.deferred = False
            self._forget(event)  # prevent tracking conflicts

            if not include_framework and isinstance(
                event,
                (PreCommitEvent, CommitEvent),
            ):
                continue

            if isinstance(event, allowed_types):
                captured.append(event)

        return _real_reemit(self)

    Framework._emit = _wrapped_emit  # type: ignore
    Framework.reemit = _wrapped_reemit  # type: ignore

    yield captured

    Framework._emit = _real_emit  # type: ignore
    Framework.reemit = _real_reemit  # type: ignore



import dataclasses
from typing import Dict, List

from scenario.state import _DCBase


@dataclasses.dataclass(frozen=True)
class Foo(_DCBase):
    a: int
    b: List[int]
    c: Dict[int, List[int]]


def test_base_case():
    l = [1, 2]
    l1 = [1, 2, 3]
    d = {1: l1}
    f = Foo(1, l, d)
    g = f.replace(a=2)

    assert g.a == 2
    assert g.b == l
    assert g.c == d
    assert g.c[1] == l1


def test_dedup_on_replace():
    l = [1, 2]
    l1 = [1, 2, 3]
    d = {1: l1}
    f = Foo(1, l, d)
    g = f.replace(a=2)

    l.append(3)
    l1.append(4)
    d[2] = "foobar"

    assert g.a == 2
    assert g.b == [1, 2]
    assert g.c == {1: [1, 2, 3]}
    assert g.c[1] == [1, 2, 3]



import pytest
from ops.charm import CharmBase
from ops.framework import Framework
from ops.framework import StoredState as ops_storedstate

from scenario.state import State, StoredState
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        META = {"name": "mycharm"}

        _read = {}
        _stored = ops_storedstate()
        _stored2 = ops_storedstate()

        def __init__(self, framework: Framework):
            super().__init__(framework)
            self._stored.set_default(foo="bar", baz={12: 142})
            self._stored2.set_default(foo="bar", baz={12: 142})
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            self._read["foo"] = self._stored.foo
            self._read["baz"] = self._stored.baz

    return MyCharm


def test_stored_state_default(mycharm):
    out = trigger(State(), "start", mycharm, meta=mycharm.META)
    assert out.stored_state[0].content == {"foo": "bar", "baz": {12: 142}}


def test_stored_state_initialized(mycharm):
    out = trigger(
        State(
            stored_state=[
                StoredState("MyCharm", name="_stored", content={"foo": "FOOX"}),
            ]
        ),
        "start",
        mycharm,
        meta=mycharm.META,
    )
    # todo: ordering is messy?
    assert out.stored_state[1].content == {"foo": "FOOX", "baz": {12: 142}}
    assert out.stored_state[0].content == {"foo": "bar", "baz": {12: 142}}



import datetime

import pytest
from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import SecretNotFoundError, SecretRotate

from scenario import Context
from scenario.state import Relation, Secret, State
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_get_secret_no_secret(mycharm):
    def post_event(charm: CharmBase):
        with pytest.raises(SecretNotFoundError):
            assert charm.model.get_secret(id="foo")
        with pytest.raises(SecretNotFoundError):
            assert charm.model.get_secret(label="foo")

    trigger(
        State(), "update_status", mycharm, meta={"name": "local"}, post_event=post_event
    )


def test_get_secret(mycharm):
    def post_event(charm: CharmBase):
        assert charm.model.get_secret(id="foo").get_content()["a"] == "b"

    trigger(
        State(secrets=[Secret(id="foo", contents={0: {"a": "b"}})]),
        "update_status",
        mycharm,
        meta={"name": "local"},
        post_event=post_event,
    )


def test_get_secret_peek_update(mycharm):
    def post_event(charm: CharmBase):
        assert charm.model.get_secret(id="foo").get_content()["a"] == "b"
        assert charm.model.get_secret(id="foo").peek_content()["a"] == "c"
        assert charm.model.get_secret(id="foo").get_content()["a"] == "b"

        assert charm.model.get_secret(id="foo").get_content(refresh=True)["a"] == "c"
        assert charm.model.get_secret(id="foo").get_content()["a"] == "c"

    trigger(
        State(
            secrets=[
                Secret(
                    id="foo",
                    contents={
                        0: {"a": "b"},
                        1: {"a": "c"},
                    },
                )
            ]
        ),
        "update_status",
        mycharm,
        meta={"name": "local"},
        post_event=post_event,
    )


def test_secret_changed_owner_evt_fails(mycharm):
    with pytest.raises(ValueError):
        _ = Secret(
            id="foo",
            contents={
                0: {"a": "b"},
                1: {"a": "c"},
            },
            owner="unit",
        ).changed_event


@pytest.mark.parametrize("evt_prefix", ("rotate", "expired", "remove"))
def test_consumer_events_failures(mycharm, evt_prefix):
    with pytest.raises(ValueError):
        _ = getattr(
            Secret(
                id="foo",
                contents={
                    0: {"a": "b"},
                    1: {"a": "c"},
                },
            ),
            evt_prefix + "_event",
        )


def test_add(mycharm):
    def post_event(charm: CharmBase):
        charm.unit.add_secret({"foo": "bar"}, label="mylabel")

    out = trigger(
        State(), "update_status", mycharm, meta={"name": "local"}, post_event=post_event
    )
    assert out.secrets
    secret = out.secrets[0]
    assert secret.contents[0] == {"foo": "bar"}
    assert secret.label == "mylabel"


def test_meta(mycharm):
    def post_event(charm: CharmBase):
        assert charm.model.get_secret(label="mylabel")

        secret = charm.model.get_secret(id="foo")
        info = secret.get_info()

        assert secret.label is None
        assert info.label == "mylabel"
        assert info.rotation == SecretRotate.HOURLY

    trigger(
        State(
            secrets=[
                Secret(
                    owner="unit",
                    id="foo",
                    label="mylabel",
                    description="foobarbaz",
                    rotate=SecretRotate.HOURLY,
                    contents={
                        0: {"a": "b"},
                    },
                )
            ]
        ),
        "update_status",
        mycharm,
        meta={"name": "local"},
        post_event=post_event,
    )


def test_meta_nonowner(mycharm):
    def post_event(charm: CharmBase):
        secret = charm.model.get_secret(id="foo")
        with pytest.raises(RuntimeError):
            info = secret.get_info()

    trigger(
        State(
            secrets=[
                Secret(
                    id="foo",
                    label="mylabel",
                    description="foobarbaz",
                    rotate=SecretRotate.HOURLY,
                    contents={
                        0: {"a": "b"},
                    },
                )
            ]
        ),
        "update_status",
        mycharm,
        meta={"name": "local"},
        post_event=post_event,
    )


@pytest.mark.parametrize("app", (True, False))
def test_grant(mycharm, app):
    def post_event(charm: CharmBase):
        secret = charm.model.get_secret(label="mylabel")
        foo = charm.model.get_relation("foo")
        if app:
            secret.grant(relation=foo)
        else:
            secret.grant(relation=foo, unit=foo.units.pop())

    out = trigger(
        State(
            relations=[Relation("foo", "remote")],
            secrets=[
                Secret(
                    owner="unit",
                    id="foo",
                    label="mylabel",
                    description="foobarbaz",
                    rotate=SecretRotate.HOURLY,
                    contents={
                        0: {"a": "b"},
                    },
                )
            ],
        ),
        "update_status",
        mycharm,
        meta={"name": "local", "requires": {"foo": {"interface": "bar"}}},
        post_event=post_event,
    )

    vals = list(out.secrets[0].remote_grants.values())
    assert vals == [{"remote"}] if app else [{"remote/0"}]


def test_update_metadata(mycharm):
    exp = datetime.datetime(2050, 12, 12)

    def post_event(charm: CharmBase):
        secret = charm.model.get_secret(label="mylabel")
        secret.set_info(
            label="babbuccia",
            description="blu",
            expire=exp,
            rotate=SecretRotate.DAILY,
        )

    out = trigger(
        State(
            secrets=[
                Secret(
                    owner="unit",
                    id="foo",
                    label="mylabel",
                    contents={
                        0: {"a": "b"},
                    },
                )
            ],
        ),
        "update_status",
        mycharm,
        meta={"name": "local"},
        post_event=post_event,
    )

    secret_out = out.secrets[0]
    assert secret_out.label == "babbuccia"
    assert secret_out.rotate == SecretRotate.DAILY
    assert secret_out.description == "blu"
    assert secret_out.expire == exp


def test_grant_nonowner(mycharm):
    def post_event(charm: CharmBase):
        secret = charm.model.get_secret(id="foo")
        with pytest.raises(RuntimeError):
            secret = charm.model.get_secret(label="mylabel")
            foo = charm.model.get_relation("foo")
            secret.grant(relation=foo)

    out = trigger(
        State(
            relations=[Relation("foo", "remote")],
            secrets=[
                Secret(
                    id="foo",
                    label="mylabel",
                    description="foobarbaz",
                    rotate=SecretRotate.HOURLY,
                    contents={
                        0: {"a": "b"},
                    },
                )
            ],
        ),
        "update_status",
        mycharm,
        meta={"name": "local", "requires": {"foo": {"interface": "bar"}}},
        post_event=post_event,
    )


class GrantingCharm(CharmBase):
    def __init__(self, *args):
        super().__init__(*args)
        self.framework.observe(self.on.start, self._on_start)

    def _on_start(self, _):
        secret = self.app.add_secret({"foo": "bar"})
        secret.grant(self.model.relations["bar"][0])


def test_grant_after_add():
    context = Context(
        GrantingCharm, meta={"name": "foo", "provides": {"bar": {"interface": "bar"}}}
    )
    state = State(relations=[Relation("bar")])
    context.run("start", state)



from unittest.mock import patch

from ops import CharmBase

from scenario import Action, Context, Event, State


class MyCharm(CharmBase):
    pass


def test_run():
    ctx = Context(MyCharm, meta={"name": "foo"})
    state = State()

    with patch.object(ctx, "_run") as p:
        ctx.run("start", state)

    assert p.called
    e = p.call_args.kwargs["event"]
    s = p.call_args.kwargs["state"]

    assert isinstance(e, Event)
    assert e.name == "start"
    assert s is state


def test_run_action():
    ctx = Context(MyCharm, meta={"name": "foo"})
    state = State()

    with patch.object(ctx, "_run_action") as p:
        ctx.run_action("do-foo", state)

    assert p.called
    a = p.call_args.kwargs["action"]
    s = p.call_args.kwargs["state"]

    assert isinstance(a, Action)
    assert a.event.name == "do_foo_action"
    assert s is state


def test_clear():
    ctx = Context(MyCharm, meta={"name": "foo"})
    state = State()

    ctx.run("start", state)
    assert ctx.emitted_events

    ctx.clear()
    assert not ctx.emitted_events  # and others...



import pytest
from ops import CharmBase

from scenario import Context, State
from scenario.state import Port


class MyCharm(CharmBase):
    META = {"name": "edgar"}


@pytest.fixture
def ctx():
    return Context(MyCharm, meta=MyCharm.META)


def test_open_port(ctx):
    def post_event(charm: CharmBase):
        charm.unit.open_port("tcp", 12)

    out = ctx.run("start", State(), post_event=post_event)
    port = out.opened_ports.pop()

    assert port.protocol == "tcp"
    assert port.port == 12


def test_close_port(ctx):
    def post_event(charm: CharmBase):
        assert charm.unit.opened_ports()
        charm.unit.close_port("tcp", 42)

    out = ctx.run("start", State(opened_ports={Port("tcp", 42)}), post_event=post_event)
    assert not out.opened_ports



import pytest

from scenario import Container, Event, Relation, Secret, State
from scenario.state import BindFailedError


def test_bind_relation():
    event = Event("foo-relation-changed")
    foo_relation = Relation("foo")
    state = State(relations=[foo_relation])
    assert event.bind(state).relation is foo_relation


def test_bind_relation_complex_name():
    event = Event("foo-bar-baz-relation-changed")
    foo_relation = Relation("foo_bar_baz")
    state = State(relations=[foo_relation])
    assert event.bind(state).relation is foo_relation


def test_bind_relation_notfound():
    event = Event("foo-relation-changed")
    state = State(relations=[])
    with pytest.raises(BindFailedError):
        event.bind(state)


def test_bind_relation_toomany(caplog):
    event = Event("foo-relation-changed")
    foo_relation = Relation("foo")
    foo_relation1 = Relation("foo")
    state = State(relations=[foo_relation, foo_relation1])
    event.bind(state)
    assert "too many relations" in caplog.text


def test_bind_secret():
    event = Event("secret-changed")
    secret = Secret("foo", {"a": "b"})
    state = State(secrets=[secret])
    assert event.bind(state).secret is secret


def test_bind_secret_notfound():
    event = Event("secret-changed")
    state = State(secrets=[])
    with pytest.raises(BindFailedError):
        event.bind(state)


def test_bind_container():
    event = Event("foo-pebble-ready")
    container = Container("foo")
    state = State(containers=[container])
    assert event.bind(state).container is container


def test_bind_container_notfound():
    event = Event("foo-pebble-ready")
    state = State(containers=[])
    with pytest.raises(BindFailedError):
        event.bind(state)



from typing import Type

import pytest
from ops.charm import CharmBase, CharmEvents, CollectStatusEvent, RelationDepartedEvent
from ops.framework import EventBase, Framework

from scenario.state import (
    PeerRelation,
    Relation,
    RelationBase,
    State,
    StateValidationError,
    SubordinateRelation,
)
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharmEvents(CharmEvents):
        @classmethod
        def define_event(cls, event_kind: str, event_type: "Type[EventBase]"):
            if getattr(cls, event_kind, None):
                delattr(cls, event_kind)
            return super().define_event(event_kind, event_type)

    class MyCharm(CharmBase):
        _call = None
        called = False
        on = MyCharmEvents()

        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            if self._call:
                MyCharm.called = True
                MyCharm._call(self, event)

    return MyCharm


def test_get_relation(mycharm):
    def pre_event(charm: CharmBase):
        assert charm.model.get_relation("foo")
        assert charm.model.get_relation("bar") is None
        assert charm.model.get_relation("qux")
        assert charm.model.get_relation("zoo") is None

    trigger(
        State(
            config={"foo": "bar"},
            leader=True,
            relations=[
                Relation(endpoint="foo", interface="foo", remote_app_name="remote"),
                Relation(endpoint="qux", interface="qux", remote_app_name="remote"),
            ],
        ),
        "start",
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
                "bar": {"interface": "bar"},
            },
            "provides": {
                "qux": {"interface": "qux"},
                "zoo": {"interface": "zoo"},
            },
        },
        config={"options": {"foo": {"type": "string"}}},
    )


@pytest.mark.parametrize(
    "evt_name", ("changed", "broken", "departed", "joined", "created")
)
def test_relation_events(mycharm, evt_name):
    relation = Relation(endpoint="foo", interface="foo", remote_app_name="remote")

    mycharm._call = lambda self, evt: None

    trigger(
        State(
            relations=[
                relation,
            ],
        ),
        getattr(relation, f"{evt_name}_event"),
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )

    assert mycharm.called


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "remote_app_name",
    ("remote", "prometheus", "aodeok123"),
)
def test_relation_events(mycharm, evt_name, remote_app_name):
    relation = Relation(
        endpoint="foo", interface="foo", remote_app_name=remote_app_name
    )

    def callback(charm: CharmBase, _):
        assert charm.model.get_relation("foo").app.name == remote_app_name

    mycharm._call = callback

    trigger(
        State(
            relations=[
                relation,
            ],
        ),
        getattr(relation, f"{evt_name}_event"),
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "remote_app_name",
    ("remote", "prometheus", "aodeok123"),
)
@pytest.mark.parametrize(
    "remote_unit_id",
    (0, 1),
)
def test_relation_events_attrs(mycharm, evt_name, remote_app_name, remote_unit_id):
    relation = Relation(
        endpoint="foo", interface="foo", remote_app_name=remote_app_name
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app
        assert event.unit
        if isinstance(event, RelationDepartedEvent):
            assert event.departing_unit

    mycharm._call = callback

    trigger(
        State(
            relations=[
                relation,
            ],
        ),
        getattr(relation, f"{evt_name}_event")(remote_unit_id=remote_unit_id),
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "remote_app_name",
    ("remote", "prometheus", "aodeok123"),
)
def test_relation_events_no_attrs(mycharm, evt_name, remote_app_name, caplog):
    relation = Relation(
        endpoint="foo",
        interface="foo",
        remote_app_name=remote_app_name,
        remote_units_data={0: {}, 1: {}},  # 2 units
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app  # that's always present
        assert event.unit
        assert (evt_name == "departed") is bool(getattr(event, "departing_unit", False))

    mycharm._call = callback

    trigger(
        State(
            relations=[
                relation,
            ],
        ),
        getattr(relation, f"{evt_name}_event"),
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )

    assert (
        "remote unit ID unset, and multiple remote unit IDs are present" in caplog.text
    )


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
def test_relation_events_no_remote_units(mycharm, evt_name, caplog):
    relation = Relation(
        endpoint="foo",
        interface="foo",
        remote_units_data={},  # no units
    )

    def callback(charm: CharmBase, event):
        if isinstance(event, CollectStatusEvent):
            return

        assert event.app  # that's always present
        assert not event.unit

    mycharm._call = callback

    trigger(
        State(
            relations=[
                relation,
            ],
        ),
        getattr(relation, f"{evt_name}_event"),
        mycharm,
        meta={
            "name": "local",
            "requires": {
                "foo": {"interface": "foo"},
            },
        },
    )

    assert "remote unit ID unset; no remote unit data present" in caplog.text


@pytest.mark.parametrize("data", (set(), {}, [], (), 1, 1.0, None, b""))
def test_relation_unit_data_bad_types(mycharm, data):
    with pytest.raises(StateValidationError):
        relation = Relation(
            endpoint="foo", interface="foo", remote_units_data={0: {"a": data}}
        )


@pytest.mark.parametrize("data", (set(), {}, [], (), 1, 1.0, None, b""))
def test_relation_app_data_bad_types(mycharm, data):
    with pytest.raises(StateValidationError):
        relation = Relation(endpoint="foo", interface="foo", local_app_data={"a": data})


@pytest.mark.parametrize(
    "evt_name",
    ("changed", "broken", "departed", "joined", "created"),
)
@pytest.mark.parametrize(
    "relation",
    (Relation("a"), PeerRelation("b"), SubordinateRelation("c")),
)
def test_relation_event_trigger(relation, evt_name, mycharm):
    meta = {
        "name": "mycharm",
        "requires": {"a": {"interface": "i1"}},
        "provides": {
            "c": {
                "interface": "i3",
                # this is a subordinate relation.
                "scope": "container",
            }
        },
        "peers": {"b": {"interface": "i2"}},
    }
    state = trigger(
        State(relations=[relation]),
        getattr(relation, evt_name + "_event"),
        mycharm,
        meta=meta,
    )


def test_trigger_sub_relation(mycharm):
    meta = {
        "name": "mycharm",
        "provides": {
            "foo": {
                "interface": "bar",
                # this is a subordinate relation.
                "scope": "container",
            }
        },
    }

    sub1 = SubordinateRelation(
        "foo", remote_unit_data={"1": "2"}, remote_app_name="primary1"
    )
    sub2 = SubordinateRelation(
        "foo", remote_unit_data={"3": "4"}, remote_app_name="primary2"
    )

    def post_event(charm: CharmBase):
        b_relations = charm.model.relations["foo"]
        assert len(b_relations) == 2
        for relation in b_relations:
            assert len(relation.units) == 1

    trigger(
        State(relations=[sub1, sub2]),
        "update-status",
        mycharm,
        meta=meta,
        post_event=post_event,
    )


def test_cannot_instantiate_relationbase():
    with pytest.raises(RuntimeError):
        RelationBase("")


def test_relation_ids():
    from scenario.state import _next_relation_id_counter

    initial_id = _next_relation_id_counter
    for i in range(10):
        rel = Relation("foo")
        assert rel.relation_id == initial_id + i



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
import datetime
import random
import shutil
from io import StringIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Set, Tuple, Union

from ops import JujuVersion, pebble
from ops.model import ModelError, RelationNotFoundError
from ops.model import Secret as Secret_Ops  # lol
from ops.model import (
    SecretInfo,
    SecretNotFoundError,
    SecretRotate,
    _format_action_result_dict,
    _ModelBackend,
)
from ops.pebble import Client, ExecError
from ops.testing import _TestingPebbleClient

from scenario.logger import logger as scenario_logger
from scenario.state import JujuLogLine, Mount, PeerRelation, Port, Storage

if TYPE_CHECKING:
    from scenario.context import Context
    from scenario.state import Container as ContainerSpec
    from scenario.state import (
        Event,
        ExecOutput,
        Relation,
        State,
        SubordinateRelation,
        _CharmSpec,
    )

logger = scenario_logger.getChild("mocking")


class ActionMissingFromContextError(Exception):
    """Raised when the user attempts to invoke action hook tools outside an action context."""

    # This is not an ops error: in ops, you'd have to go exceptionally out of your way to trigger
    # this flow.


class _MockExecProcess:
    def __init__(self, command: Tuple[str], change_id: int, out: "ExecOutput"):
        self._command = command
        self._change_id = change_id
        self._out = out
        self._waited = False
        self.stdout = StringIO(self._out.stdout)
        self.stderr = StringIO(self._out.stderr)

    def wait(self):
        self._waited = True
        exit_code = self._out.return_code
        if exit_code != 0:
            raise ExecError(list(self._command), exit_code, None, None)

    def wait_output(self):
        out = self._out
        exit_code = out.return_code
        if exit_code != 0:
            raise ExecError(list(self._command), exit_code, None, None)
        return out.stdout, out.stderr

    def send_signal(self, sig: Union[int, str]):  # noqa: U100
        raise NotImplementedError()


_NOT_GIVEN = object()  # non-None default value sentinel


class _MockModelBackend(_ModelBackend):
    def __init__(
        self,
        state: "State",
        event: "Event",
        charm_spec: "_CharmSpec",
        context: "Context",
    ):
        super().__init__()
        self._state = state
        self._event = event
        self._context = context
        self._charm_spec = charm_spec

    def opened_ports(self) -> Set[Port]:
        return set(self._state.opened_ports)

    def open_port(self, protocol: str, port: Optional[int] = None):
        # fixme: the charm will get hit with a StateValidationError
        #  here, not the expected ModelError...
        port = Port(protocol, port)
        ports = self._state.opened_ports
        if port not in ports:
            ports.append(port)

    def close_port(self, protocol: str, port: Optional[int] = None):
        port = Port(protocol, port)
        ports = self._state.opened_ports
        if port in ports:
            ports.remove(port)

    def get_pebble(self, socket_path: str) -> "Client":
        container_name = socket_path.split("/")[
            3
        ]  # /charm/containers/<container_name>/pebble.socket
        container_root = self._context._get_container_root(container_name)
        try:
            mounts = self._state.get_container(container_name).mounts
        except ValueError:
            # container not defined in state.
            mounts = {}

        return _MockPebbleClient(
            socket_path=socket_path,
            container_root=container_root,
            mounts=mounts,
            state=self._state,
            event=self._event,
            charm_spec=self._charm_spec,
        )

    def _get_relation_by_id(
        self,
        rel_id,
    ) -> Union["Relation", "SubordinateRelation", "PeerRelation"]:
        try:
            return next(
                filter(lambda r: r.relation_id == rel_id, self._state.relations),
            )
        except StopIteration:
            raise RelationNotFoundError()

    def _get_secret(self, id=None, label=None):
        canonicalize_id = Secret_Ops._canonicalize_id

        if id:
            # in scenario, you can create Secret(id="foo"),
            # but ops.Secret will prepend a "secret:" prefix to that ID.
            # we allow getting secret by either version.
            try:
                return next(
                    filter(
                        lambda s: canonicalize_id(s.id) == canonicalize_id(id),
                        self._state.secrets,
                    ),
                )
            except StopIteration:
                raise SecretNotFoundError()
        elif label:
            try:
                return next(filter(lambda s: s.label == label, self._state.secrets))
            except StopIteration:
                raise SecretNotFoundError()
        else:
            # if all goes well, this should never be reached. ops.model.Secret will check upon
            # instantiation that either an id or a label are set, and raise a TypeError if not.
            raise RuntimeError("need id or label.")

    @staticmethod
    def _generate_secret_id():
        id = "".join(map(str, [random.choice(list(range(10))) for _ in range(20)]))
        return f"secret:{id}"

    def _check_app_data_access(self, is_app: bool):
        if not isinstance(is_app, bool):
            raise TypeError("is_app parameter to relation_get must be a boolean")

        if not is_app:
            return

        version = JujuVersion(self._context.juju_version)
        if not version.has_app_data():
            raise RuntimeError(
                f"setting application data is not supported on Juju version {version}",
            )

    def relation_get(self, relation_id: int, member_name: str, is_app: bool):
        self._check_app_data_access(is_app)
        relation = self._get_relation_by_id(relation_id)
        if is_app and member_name == self.app_name:
            return relation.local_app_data
        elif is_app:
            return relation.remote_app_data
        elif member_name == self.unit_name:
            return relation.local_unit_data

        unit_id = int(member_name.split("/")[-1])
        return relation._get_databag_for_remote(unit_id)  # noqa

    def is_leader(self):
        return self._state.leader

    def status_get(self, *, is_app: bool = False):
        status, message = self._state.app_status if is_app else self._state.unit_status
        return {"status": status, "message": message}

    def relation_ids(self, relation_name):
        return [
            rel.relation_id
            for rel in self._state.relations
            if rel.endpoint == relation_name
        ]

    def relation_list(self, relation_id: int) -> Tuple[str]:
        relation = self._get_relation_by_id(relation_id)

        if isinstance(relation, PeerRelation):
            return tuple(
                f"{self.app_name}/{unit_id}" for unit_id in relation.peers_data
            )
        return tuple(
            f"{relation.remote_app_name}/{unit_id}"
            for unit_id in relation._remote_unit_ids
        )

    def config_get(self):
        state_config = self._state.config

        # add defaults
        charm_config = self._charm_spec.config
        if not charm_config:
            return state_config

        for key, value in charm_config["options"].items():
            # if it has a default, and it's not overwritten from State, use it:
            if key not in state_config:
                default_value = value.get("default", _NOT_GIVEN)
                if default_value is not _NOT_GIVEN:  # accept False as default value
                    state_config[key] = default_value

        return state_config  # full config

    def network_get(self, binding_name: str, relation_id: Optional[int] = None):
        if relation_id:
            logger.warning("network-get -r not implemented")

        relations = self._state.get_relations(binding_name)
        if not relations:
            raise RelationNotFoundError()

        network = next(filter(lambda r: r.name == binding_name, self._state.networks))
        return network.hook_tool_output_fmt()

    # setter methods: these can mutate the state.
    def application_version_set(self, version: str):
        if workload_version := self._state.workload_version:
            # do not record if empty = unset
            self._context.workload_version_history.append(workload_version)

        self._state._update_workload_version(version)

    def status_set(self, status: str, message: str = "", *, is_app: bool = False):
        self._context._record_status(self._state, is_app)
        self._state._update_status(status, message, is_app)

    def juju_log(self, level: str, message: str):
        self._context.juju_log.append(JujuLogLine(level, message))

    def relation_set(self, relation_id: int, key: str, value: str, is_app: bool):
        self._check_app_data_access(is_app)
        relation = self._get_relation_by_id(relation_id)
        if is_app:
            if not self._state.leader:
                # will in practice not be reached because RelationData will check leadership
                # and raise RelationDataAccessError upstream on this path
                raise RuntimeError("needs leadership to set app data")
            tgt = relation.local_app_data
        else:
            tgt = relation.local_unit_data
        tgt[key] = value
        return None

    def secret_add(
        self,
        content: Dict[str, str],
        *,
        label: Optional[str] = None,
        description: Optional[str] = None,
        expire: Optional[datetime.datetime] = None,
        rotate: Optional[SecretRotate] = None,
        owner: Optional[str] = None,
    ) -> str:
        from scenario.state import Secret

        id = self._generate_secret_id()
        secret = Secret(
            id=id,
            contents={0: content},
            label=label,
            description=description,
            expire=expire,
            rotate=rotate,
            owner=owner,
        )
        self._state.secrets.append(secret)
        return id

    def secret_get(
        self,
        *,
        id: str = None,
        label: str = None,
        refresh: bool = False,
        peek: bool = False,
    ) -> Dict[str, str]:
        secret = self._get_secret(id, label)
        revision = secret.revision
        if peek or refresh:
            revision = max(secret.contents.keys())
            if refresh:
                secret._set_revision(revision)

        return secret.contents[revision]

    def secret_info_get(
        self,
        *,
        id: Optional[str] = None,
        label: Optional[str] = None,
    ) -> SecretInfo:
        secret = self._get_secret(id, label)
        if not secret.owner:
            raise RuntimeError(f"not the owner of {secret}")

        return SecretInfo(
            id=secret.id,
            label=secret.label,
            revision=max(secret.contents),
            expires=secret.expire,
            rotation=secret.rotate,
            rotates=None,  # not implemented yet.
        )

    def secret_set(
        self,
        id: str,
        *,
        content: Optional[Dict[str, str]] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        expire: Optional[datetime.datetime] = None,
        rotate: Optional[SecretRotate] = None,
    ):
        secret = self._get_secret(id, label)
        if not secret.owner:
            raise RuntimeError(f"not the owner of {secret}")

        secret._update_metadata(
            content=content,
            label=label,
            description=description,
            expire=expire,
            rotate=rotate,
        )

    def secret_grant(self, id: str, relation_id: int, *, unit: Optional[str] = None):
        secret = self._get_secret(id)
        if not secret.owner:
            raise RuntimeError(f"not the owner of {secret}")

        grantee = unit or self._get_relation_by_id(relation_id).remote_app_name

        if not secret.remote_grants.get(relation_id):
            secret.remote_grants[relation_id] = set()

        secret.remote_grants[relation_id].add(grantee)

    def secret_revoke(self, id: str, relation_id: int, *, unit: Optional[str] = None):
        secret = self._get_secret(id)
        if not secret.owner:
            raise RuntimeError(f"not the owner of {secret}")

        grantee = unit or self._get_relation_by_id(relation_id).remote_app_name
        secret.remote_grants[relation_id].remove(grantee)

    def secret_remove(self, id: str, *, revision: Optional[int] = None):
        secret = self._get_secret(id)
        if not secret.owner:
            raise RuntimeError(f"not the owner of {secret}")

        if revision:
            del secret.contents[revision]
        else:
            secret.contents.clear()

    def relation_remote_app_name(self, relation_id: int) -> Optional[str]:
        # ops catches relationnotfounderrors and returns None:
        try:
            relation = self._get_relation_by_id(relation_id)
        except RelationNotFoundError:
            return None
        return relation.remote_app_name

    def action_set(self, results: Dict[str, Any]):
        if not self._event.action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-set",
            )
        # let ops validate the results dict
        _format_action_result_dict(results)
        # but then we will store it in its unformatted,
        # original form for testing ease
        self._context._action_results = results

    def action_fail(self, message: str = ""):
        if not self._event.action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-fail",
            )
        self._context._action_failure = message

    def action_log(self, message: str):
        if not self._event.action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-log",
            )
        self._context._action_logs.append(message)

    def action_get(self):
        action = self._event.action
        if not action:
            raise ActionMissingFromContextError(
                "not in the context of an action event: cannot action-get",
            )
        return action.params

    def storage_add(self, name: str, count: int = 1):
        if not isinstance(count, int) or isinstance(count, bool):
            raise TypeError(
                f"storage count must be integer, got: {count} ({type(count)})",
            )

        if "/" in name:
            # this error is raised by ops.testing but not by ops at runtime
            raise ModelError('storage name cannot contain "/"')

        self._context.requested_storages[name] = count

    def storage_list(self, name: str) -> List[int]:
        return [
            storage.index for storage in self._state.storage if storage.name == name
        ]

    def _storage_event_details(self) -> Tuple[int, str]:
        storage = self._event.storage
        if not storage:
            # only occurs if this method is called when outside the scope of a storage event
            raise RuntimeError('unable to find storage key in ""')
        fs_path = storage.get_filesystem(self._context)
        return storage.index, str(fs_path)

    def storage_get(self, storage_name_id: str, attribute: str) -> str:
        if not len(attribute) > 0:  # assume it's an empty string.
            raise RuntimeError(
                'calling storage_get with `attribute=""` will return a dict '
                "and not a string. This usage is not supported.",
            )

        if attribute != "location":
            # this should not happen: in ops it's hardcoded to be "location"
            raise NotImplementedError(
                f"storage-get not implemented for attribute={attribute}",
            )

        name, index = storage_name_id.split("/")
        index = int(index)
        storages: List[Storage] = [
            s for s in self._state.storage if s.name == name and s.index == index
        ]

        # should not really happen: sanity checks. In practice, ops will guard against these paths.
        if not storages:
            raise RuntimeError(f"Storage with name={name} and index={index} not found.")
        if len(storages) > 1:
            raise RuntimeError(
                f"Multiple Storage instances with name={name} and index={index} found. "
                f"Inconsistent state.",
            )

        storage = storages[0]
        fs_path = storage.get_filesystem(self._context)
        return str(fs_path)

    def planned_units(self) -> int:
        return self._state.planned_units

    # legacy ops API that we don't intend to mock:
    def pod_spec_set(
        self,
        spec: Mapping[str, Any],  # noqa: U100
        k8s_resources: Optional[Mapping[str, Any]] = None,  # noqa: U100
    ):
        raise NotImplementedError(
            "pod-spec-set is not implemented in Scenario (and probably never will be: "
            "it's deprecated API)",
        )

    def add_metrics(
        self,
        metrics: Mapping[str, Union[int, float]],  # noqa: U100
        labels: Optional[Mapping[str, str]] = None,  # noqa: U100
    ) -> None:
        raise NotImplementedError(
            "add-metrics is not implemented in Scenario (and probably never will be: "
            "it's deprecated API)",
        )

    def resource_get(self, resource_name: str) -> str:
        try:
            return str(self._state.resources[resource_name])
        except KeyError:
            # ops will not let us get there if the resource name is unknown from metadata.
            # but if the user forgot to add it in State, then we remind you of that.
            raise RuntimeError(
                f"Inconsistent state: "
                f"resource {resource_name} not found in State. please pass it.",
            )


class _MockPebbleClient(_TestingPebbleClient):
    def __init__(
        self,
        socket_path: str,
        container_root: Path,
        mounts: Dict[str, Mount],
        *,
        state: "State",
        event: "Event",
        charm_spec: "_CharmSpec",
    ):
        self._state = state
        self.socket_path = socket_path
        self._event = event
        self._charm_spec = charm_spec

        # wipe just in case
        if container_root.exists():
            # Path.rmdir will fail if root is nonempty
            shutil.rmtree(container_root)

        # initialize simulated filesystem
        container_root.mkdir(parents=True)
        for _, mount in mounts.items():
            mounting_dir = container_root / mount.location[1:]
            mounting_dir.parent.mkdir(parents=True, exist_ok=True)
            mounting_dir.symlink_to(mount.src)

        self._root = container_root

    def get_plan(self) -> pebble.Plan:
        return self._container.plan

    @property
    def _container(self) -> "ContainerSpec":
        container_name = self.socket_path.split("/")[-2]
        try:
            return next(
                filter(lambda x: x.name == container_name, self._state.containers),
            )
        except StopIteration:
            raise RuntimeError(
                f"container with name={container_name!r} not found. "
                f"Did you forget a Container, or is the socket path "
                f"{self.socket_path!r} wrong?",
            )

    @property
    def _layers(self) -> Dict[str, pebble.Layer]:
        return self._container.layers

    @property
    def _service_status(self) -> Dict[str, pebble.ServiceStatus]:
        return self._container.service_status

    def exec(self, *args, **kwargs):  # noqa: U100
        cmd = tuple(args[0])
        out = self._container.exec_mock.get(cmd)
        if not out:
            raise RuntimeError(
                f"mock for cmd {cmd} not found. Please pass to the Container "
                f"{self._container.name} a scenario.ExecOutput mock for the "
                f"command your charm is attempting to run, or patch "
                f"out whatever leads to the call.",
            )

        change_id = out._run()
        return _MockExecProcess(change_id=change_id, command=cmd, out=out)

    def _check_connection(self):
        if not self._container.can_connect:  # pyright: reportPrivateUsage=false
            msg = (
                f"Cannot connect to Pebble; did you forget to set "
                f"can_connect=True for container {self._container.name}?"
            )
            raise pebble.ConnectionError(msg)



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
import inspect
import os
from typing import TYPE_CHECKING, Any, Sequence

import ops.charm
import ops.framework
import ops.model
import ops.storage
from ops import CharmBase
from ops.charm import CharmMeta
from ops.log import setup_root_logging

# use logger from ops.main so that juju_log will be triggered
from ops.main import CHARM_STATE_FILE, _Dispatcher, _get_charm_dir, _get_event_args
from ops.main import logger as ops_logger

if TYPE_CHECKING:
    from scenario.context import Context
    from scenario.state import Event, State, _CharmSpec


class NoObserverError(RuntimeError):
    """Error raised when the event being dispatched has no registered observers."""


class BadOwnerPath(RuntimeError):
    """Error raised when the owner path does not lead to a valid ObjectEvents instance."""


def _get_owner(root: Any, path: Sequence[str]) -> ops.ObjectEvents:
    """Walk path on root to an ObjectEvents instance."""
    obj = root
    for step in path:
        try:
            obj = getattr(obj, step)
        except AttributeError:
            raise BadOwnerPath(
                f"event_owner_path {path!r} invalid: {step!r} leads to nowhere.",
            )
    if not isinstance(obj, ops.ObjectEvents):
        raise BadOwnerPath(
            f"event_owner_path {path!r} invalid: does not lead to "
            f"an ObjectEvents instance.",
        )
    return obj


def _emit_charm_event(
    charm: "CharmBase",
    event_name: str,
    event: "Event" = None,
):
    """Emits a charm event based on a Juju event name.

    Args:
        charm: A charm instance to emit an event from.
        event_name: A Juju event name to emit on a charm.
        event_owner_path: Event source lookup path.
    """
    owner = _get_owner(charm, event.owner_path) if event else charm.on

    try:
        event_to_emit = getattr(owner, event_name)
    except AttributeError:
        ops_logger.debug("Event %s not defined for %s.", event_name, charm)
        raise NoObserverError(
            f"Cannot fire {event_name!r} on {owner}: "
            f"invalid event (not on charm.on). "
            f"Use Context.run_custom instead.",
        )

    args, kwargs = _get_event_args(charm, event_to_emit)
    ops_logger.debug("Emitting Juju event %s.", event_name)
    event_to_emit.emit(*args, **kwargs)


def setup_framework(
    charm_dir,
    state: "State",
    event: "Event",
    context: "Context",
    charm_spec: "_CharmSpec",
):
    from scenario.mocking import _MockModelBackend

    model_backend = _MockModelBackend(  # pyright: reportPrivateUsage=false
        state=state,
        event=event,
        context=context,
        charm_spec=charm_spec,
    )
    debug = "JUJU_DEBUG" in os.environ
    setup_root_logging(model_backend, debug=debug)
    ops_logger.debug(
        "Operator Framework %s up and running.",
        ops.__version__,
    )  # type:ignore

    metadata = (charm_dir / "metadata.yaml").read_text()
    actions_meta = charm_dir / "actions.yaml"
    if actions_meta.exists():
        actions_metadata = actions_meta.read_text()
    else:
        actions_metadata = None

    meta = CharmMeta.from_yaml(metadata, actions_metadata)
    model = ops.model.Model(meta, model_backend)

    charm_state_path = charm_dir / CHARM_STATE_FILE

    # TODO: add use_juju_for_storage support
    store = ops.storage.SQLiteStorage(charm_state_path)
    framework = ops.framework.Framework(store, charm_dir, meta, model)
    framework.set_breakpointhook()
    return framework


def setup_charm(charm_class, framework, dispatcher):
    sig = inspect.signature(charm_class)
    sig.bind(framework)  # signature check

    charm = charm_class(framework)
    dispatcher.ensure_event_links(charm)
    return charm


def setup(state: "State", event: "Event", context: "Context", charm_spec: "_CharmSpec"):
    """Setup dispatcher, framework and charm objects."""
    charm_class = charm_spec.charm_type
    charm_dir = _get_charm_dir()

    dispatcher = _Dispatcher(charm_dir)
    dispatcher.run_any_legacy_hook()

    framework = setup_framework(charm_dir, state, event, context, charm_spec)
    charm = setup_charm(charm_class, framework, dispatcher)
    return dispatcher, framework, charm


class Ops:
    """Class to manage stepping through ops setup, event emission and framework commit."""

    def __init__(
        self,
        state: "State",
        event: "Event",
        context: "Context",
        charm_spec: "_CharmSpec",
    ):
        self.state = state
        self.event = event
        self.context = context
        self.charm_spec = charm_spec

        # set by setup()
        self.dispatcher = None
        self.framework = None
        self.charm = None

        self._has_setup = False
        self._has_emitted = False
        self._has_committed = False

    def setup(self):
        """Setup framework, charm and dispatcher."""
        self._has_setup = True
        self.dispatcher, self.framework, self.charm = setup(
            self.state,
            self.event,
            self.context,
            self.charm_spec,
        )

    def emit(self):
        """Emit the event on the charm."""
        if not self._has_setup:
            raise RuntimeError("should .setup() before you .emit()")
        self._has_emitted = True

        try:
            if not self.dispatcher.is_restricted_context():
                self.framework.reemit()

            _emit_charm_event(self.charm, self.dispatcher.event_name, self.event)

        except Exception:
            self.framework.close()
            raise

    def commit(self):
        """Commit the framework and teardown."""
        if not self._has_emitted:
            raise RuntimeError("should .emit() before you .commit()")

        # emit collect-status events
        ops.charm._evaluate_status(self.charm)

        self._has_committed = True

        try:
            self.framework.commit()
        finally:
            self.framework.close()

    def finalize(self):
        """Step through all non-manually-called procedures and run them."""
        if not self._has_setup:
            self.setup()
        if not self._has_emitted:
            self.emit()
        if not self._has_committed:
            self.commit()



import pytest
from ops.charm import CharmBase, CharmEvents, CollectStatusEvent, StartEvent
from ops.framework import CommitEvent, EventBase, EventSource, PreCommitEvent

from scenario import Event, State
from scenario.capture_events import capture_events
from tests.helpers import trigger


class Foo(EventBase):
    pass


class MyCharmEvents(CharmEvents):
    foo = EventSource(Foo)


class MyCharm(CharmBase):
    META = {"name": "mycharm"}
    on = MyCharmEvents()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.framework.observe(self.on.start, self._on_start)
        self.framework.observe(self.on.foo, self._on_foo)

    def _on_start(self, e):
        self.on.foo.emit()

    def _on_foo(self, e):
        pass


def test_capture_custom_evt():
    with capture_events(Foo) as emitted:
        trigger(State(), "foo", MyCharm, meta=MyCharm.META)

    assert len(emitted) == 1
    assert isinstance(emitted[0], Foo)


def test_capture_custom_evt_nonspecific_capture():
    with capture_events() as emitted:
        trigger(State(), "foo", MyCharm, meta=MyCharm.META)

    assert len(emitted) == 1
    assert isinstance(emitted[0], Foo)


def test_capture_custom_evt_nonspecific_capture_include_fw_evts():
    with capture_events(include_framework=True) as emitted:
        trigger(State(), "foo", MyCharm, meta=MyCharm.META)

    assert len(emitted) == 4
    assert isinstance(emitted[0], Foo)
    assert isinstance(emitted[1], CollectStatusEvent)
    assert isinstance(emitted[2], PreCommitEvent)
    assert isinstance(emitted[3], CommitEvent)


def test_capture_juju_evt():
    with capture_events() as emitted:
        trigger(State(), "start", MyCharm, meta=MyCharm.META)

    assert len(emitted) == 2
    assert isinstance(emitted[0], StartEvent)
    assert isinstance(emitted[1], Foo)


def test_capture_deferred_evt():
    # todo: this test should pass with ops < 2.1 as well
    with capture_events() as emitted:
        trigger(
            State(deferred=[Event("foo").deferred(handler=MyCharm._on_foo)]),
            "start",
            MyCharm,
            meta=MyCharm.META,
        )

    assert len(emitted) == 3
    assert isinstance(emitted[0], Foo)
    assert isinstance(emitted[1], StartEvent)
    assert isinstance(emitted[2], Foo)


def test_capture_no_deferred_evt():
    # todo: this test should pass with ops < 2.1 as well
    with capture_events(include_deferred=False) as emitted:
        trigger(
            State(deferred=[Event("foo").deferred(handler=MyCharm._on_foo)]),
            "start",
            MyCharm,
            meta=MyCharm.META,
        )

    assert len(emitted) == 2
    assert isinstance(emitted[0], StartEvent)
    assert isinstance(emitted[1], Foo)



import pytest
from ops import CharmBase, ModelError

from scenario import Context, State, Storage


class MyCharmWithStorage(CharmBase):
    META = {"name": "charlene", "storage": {"foo": {"type": "filesystem"}}}


class MyCharmWithoutStorage(CharmBase):
    META = {"name": "patrick"}


@pytest.fixture
def storage_ctx():
    return Context(MyCharmWithStorage, meta=MyCharmWithStorage.META)


@pytest.fixture
def no_storage_ctx():
    return Context(MyCharmWithoutStorage, meta=MyCharmWithoutStorage.META)


def test_storage_get_null(no_storage_ctx):
    with no_storage_ctx.manager("update-status", State()) as mgr:
        storages = mgr.charm.model.storages
        assert not len(storages)


def test_storage_get_unknown_name(storage_ctx):
    with storage_ctx.manager("update-status", State()) as mgr:
        storages = mgr.charm.model.storages
        # not in metadata
        with pytest.raises(KeyError):
            storages["bar"]


def test_storage_request_unknown_name(storage_ctx):
    with storage_ctx.manager("update-status", State()) as mgr:
        storages = mgr.charm.model.storages
        # not in metadata
        with pytest.raises(ModelError):
            storages.request("bar")


def test_storage_get_some(storage_ctx):
    with storage_ctx.manager("update-status", State()) as mgr:
        storages = mgr.charm.model.storages
        # known but none attached
        assert storages["foo"] == []


@pytest.mark.parametrize("n", (1, 3, 5))
def test_storage_add(storage_ctx, n):
    with storage_ctx.manager("update-status", State()) as mgr:
        storages = mgr.charm.model.storages
        storages.request("foo", n)

    assert storage_ctx.requested_storages["foo"] == n


def test_storage_usage(storage_ctx):
    storage = Storage("foo")
    # setup storage with some content
    (storage.get_filesystem(storage_ctx) / "myfile.txt").write_text("helloworld")

    with storage_ctx.manager("update-status", State(storage=[storage])) as mgr:
        foo = mgr.charm.model.storages["foo"][0]
        loc = foo.location
        path = loc / "myfile.txt"
        assert path.exists()
        assert path.read_text() == "helloworld"

        myfile = loc / "path.py"
        myfile.write_text("helloworlds")

    # post-mortem: inspect fs contents.
    assert (
        storage.get_filesystem(storage_ctx) / "path.py"
    ).read_text() == "helloworlds"


def test_storage_attached_event(storage_ctx):
    storage = Storage("foo")
    storage_ctx.run(storage.attached_event, State(storage=[storage]))


def test_storage_detaching_event(storage_ctx):
    storage = Storage("foo")
    storage_ctx.run(storage.detaching_event, State(storage=[storage]))



import pytest
from ops.charm import CharmBase
from ops.framework import Framework

from scenario.state import Event, Network, Relation, State, _CharmSpec
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_config_get(mycharm):
    def check_cfg(charm: CharmBase):
        assert charm.config["foo"] == "bar"
        assert charm.config["baz"] == 1

    trigger(
        State(
            config={"foo": "bar", "baz": 1},
        ),
        "update_status",
        mycharm,
        meta={"name": "foo"},
        config={"options": {"foo": {"type": "string"}, "baz": {"type": "integer"}}},
        post_event=check_cfg,
    )


def test_config_get_default_from_meta(mycharm):
    def check_cfg(charm: CharmBase):
        assert charm.config["foo"] == "bar"
        assert charm.config["baz"] == 2
        assert charm.config["qux"] is False

    trigger(
        State(
            config={"foo": "bar"},
        ),
        "update_status",
        mycharm,
        meta={"name": "foo"},
        config={
            "options": {
                "foo": {"type": "string"},
                "baz": {"type": "integer", "default": 2},
                "qux": {"type": "boolean", "default": False},
            },
        },
        post_event=check_cfg,
    )



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
import marshal
import os
import re
import tempfile
import typing
from contextlib import contextmanager
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    ContextManager,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import yaml
from ops.framework import EventBase, _event_regex
from ops.storage import NoSnapshotError, SQLiteStorage

from scenario.capture_events import capture_events
from scenario.logger import logger as scenario_logger
from scenario.ops_main_mock import NoObserverError
from scenario.state import DeferredEvent, PeerRelation, StoredState

if TYPE_CHECKING:
    from ops.testing import CharmType

    from scenario.context import Context
    from scenario.ops_main_mock import Ops
    from scenario.state import AnyRelation, Event, State, _CharmSpec

    _CT = TypeVar("_CT", bound=Type[CharmType])

    PathLike = Union[str, Path]

logger = scenario_logger.getChild("runtime")
STORED_STATE_REGEX = re.compile(
    r"((?P<owner_path>.*)\/)?(?P<data_type_name>\D+)\[(?P<name>.*)\]",
)
EVENT_REGEX = re.compile(_event_regex)

RUNTIME_MODULE = Path(__file__).parent


class ScenarioRuntimeError(RuntimeError):
    """Base class for exceptions raised by scenario.runtime."""


class UncaughtCharmError(ScenarioRuntimeError):
    """Error raised if the charm raises while handling the event being dispatched."""


class InconsistentScenarioError(ScenarioRuntimeError):
    """Error raised when the combination of state and event is inconsistent."""


class UnitStateDB:
    """Represents the unit-state.db."""

    def __init__(self, db_path: Union[Path, str]):
        self._db_path = db_path
        self._state_file = Path(self._db_path)

    def _open_db(self) -> Optional[SQLiteStorage]:
        """Open the db."""
        return SQLiteStorage(self._state_file)

    def get_stored_state(self) -> List["StoredState"]:
        """Load any StoredState data structures from the db."""

        db = self._open_db()

        stored_state = []
        for handle_path in db.list_snapshots():
            if not EVENT_REGEX.match(handle_path) and (
                match := STORED_STATE_REGEX.match(handle_path)
            ):
                stored_state_snapshot = db.load_snapshot(handle_path)
                kwargs = match.groupdict()
                sst = StoredState(content=stored_state_snapshot, **kwargs)
                stored_state.append(sst)

        db.close()
        return stored_state

    def get_deferred_events(self) -> List["DeferredEvent"]:
        """Load any DeferredEvent data structures from the db."""

        db = self._open_db()

        deferred = []
        for handle_path in db.list_snapshots():
            if EVENT_REGEX.match(handle_path):
                notices = db.notices(handle_path)
                for handle, owner, observer in notices:
                    try:
                        snapshot_data = db.load_snapshot(handle)
                    except NoSnapshotError:
                        snapshot_data = {}

                    event = DeferredEvent(
                        handle_path=handle,
                        owner=owner,
                        observer=observer,
                        snapshot_data=snapshot_data,
                    )
                    deferred.append(event)

        db.close()
        return deferred

    def apply_state(self, state: "State"):
        """Add DeferredEvent and StoredState from this State instance to the storage."""
        db = self._open_db()
        for event in state.deferred:
            db.save_notice(event.handle_path, event.owner, event.observer)
            try:
                marshal.dumps(event.snapshot_data)
            except ValueError as e:
                raise ValueError(
                    f"unable to save the data for {event}, it must contain only simple types.",
                ) from e
            db.save_snapshot(event.handle_path, event.snapshot_data)

        for stored_state in state.stored_state:
            db.save_snapshot(stored_state.handle_path, stored_state.content)

        db.close()


class _OpsMainContext:
    """Context manager representing ops.main execution context.

    When entered, ops.main sets up everything up until the charm.
    When .emit() is called, ops.main proceeds with emitting the event.
    When exited, if .emit has not been called manually, it is called automatically.
    """

    def __init__(self):
        self._has_emitted = False

    def __enter__(self):
        pass

    def emit(self):
        self._has_emitted = True

    def __exit__(self, exc_type, exc_val, exc_tb):  # noqa: U100
        if not self._has_emitted:
            self.emit()


class Runtime:
    """Charm runtime wrapper.

    This object bridges a local environment and a charm artifact.
    """

    def __init__(
        self,
        charm_spec: "_CharmSpec",
        charm_root: Optional["PathLike"] = None,
        juju_version: str = "3.0.0",
    ):
        self._charm_spec = charm_spec
        self._juju_version = juju_version
        self._charm_root = charm_root

        app_name = self._charm_spec.meta.get("name")
        if not app_name:
            raise ValueError('invalid metadata: mandatory "name" field is missing.')

        self._app_name = app_name

    @staticmethod
    def _cleanup_env(env):
        # TODO consider cleaning up env on __delete__, but ideally you should be
        #  running this in a clean env or a container anyway.
        # cleanup the env, in case we'll be firing multiple events, we don't want to pollute it.
        for key in env:
            # os.unsetenv does not always seem to work !?
            del os.environ[key]

    def _get_event_env(self, state: "State", event: "Event", charm_root: Path):
        """Build the simulated environment the operator framework expects."""
        if event.name.endswith("_action"):
            # todo: do we need some special metadata, or can we assume action names
            #  are always dashes?
            action_name = event.name[: -len("_action")].replace("_", "-")
        else:
            action_name = ""

        env = {
            "JUJU_VERSION": self._juju_version,
            "JUJU_UNIT_NAME": f"{self._app_name}/{state.unit_id}",
            "_": "./dispatch",
            "JUJU_DISPATCH_PATH": f"hooks/{event.name}",
            "JUJU_MODEL_NAME": state.model.name,
            "JUJU_ACTION_NAME": action_name,
            "JUJU_MODEL_UUID": state.model.uuid,
            "JUJU_CHARM_DIR": str(charm_root.absolute()),
            # todo consider setting pwd, (python)path
        }

        relation: "AnyRelation"

        if event._is_relation_event and (relation := event.relation):
            if isinstance(relation, PeerRelation):
                remote_app_name = self._app_name
            else:
                remote_app_name = relation.remote_app_name
            env.update(
                {
                    "JUJU_RELATION": relation.endpoint,
                    "JUJU_RELATION_ID": str(relation.relation_id),
                    "JUJU_REMOTE_APP": remote_app_name,
                },
            )

            remote_unit_id = event.relation_remote_unit_id

            # don't check truthiness because remote_unit_id could be 0
            if remote_unit_id is None:
                remote_unit_ids = relation._remote_unit_ids  # pyright: ignore

                if len(remote_unit_ids) == 1:
                    remote_unit_id = remote_unit_ids[0]
                    logger.info(
                        "there's only one remote unit, so we set JUJU_REMOTE_UNIT to it, "
                        "but you probably should be parametrizing the event with `remote_unit_id` "
                        "to be explicit.",
                    )
                elif len(remote_unit_ids) > 1:
                    remote_unit_id = remote_unit_ids[0]
                    logger.warning(
                        "remote unit ID unset, and multiple remote unit IDs are present; "
                        "We will pick the first one and hope for the best. You should be passing "
                        "`remote_unit_id` to the Event constructor.",
                    )
                else:
                    logger.warning(
                        "remote unit ID unset; no remote unit data present. "
                        "Is this a realistic scenario?",  # TODO: is it?
                    )

            if remote_unit_id is not None:
                remote_unit = f"{remote_app_name}/{remote_unit_id}"
                env["JUJU_REMOTE_UNIT"] = remote_unit
                if event.name.endswith("_relation_departed"):
                    env["JUJU_DEPARTING_UNIT"] = remote_unit

        if container := event.container:
            env.update({"JUJU_WORKLOAD_NAME": container.name})

        if storage := event.storage:
            env.update({"JUJU_STORAGE_ID": f"{storage.name}/{storage.index}"})

        if secret := event.secret:
            env.update(
                {
                    "JUJU_SECRET_ID": secret.id,
                    "JUJU_SECRET_LABEL": secret.label or "",
                },
            )

        return env

    @staticmethod
    def _wrap(charm_type: "_CT") -> "_CT":
        # dark sorcery to work around framework using class attrs to hold on to event sources
        # todo this should only be needed if we call play multiple times on the same runtime.
        #  can we avoid it?
        class WrappedEvents(charm_type.on.__class__):
            pass

        WrappedEvents.__name__ = charm_type.on.__class__.__name__

        class WrappedCharm(charm_type):  # type: ignore
            on = WrappedEvents()

        WrappedCharm.__name__ = charm_type.__name__
        return WrappedCharm

    @contextmanager
    def _virtual_charm_root(self) -> typing.ContextManager[Path]:
        # If we are using runtime on a real charm, we can make some assumptions about the
        # directory structure we are going to find.
        # If we're, say, dynamically defining charm types and doing tests on them, we'll have to
        # generate the metadata files ourselves. To be sure, we ALWAYS use a tempdir. Ground truth
        # is what the user passed via the CharmSpec
        spec = self._charm_spec

        if charm_virtual_root := self._charm_root:
            charm_virtual_root_is_custom = True
            virtual_charm_root = Path(charm_virtual_root)
        else:
            charm_virtual_root = tempfile.TemporaryDirectory()
            virtual_charm_root = Path(charm_virtual_root.name)
            charm_virtual_root_is_custom = False

        metadata_yaml = virtual_charm_root / "metadata.yaml"
        config_yaml = virtual_charm_root / "config.yaml"
        actions_yaml = virtual_charm_root / "actions.yaml"

        metadata_files_present: Dict[Path, Optional[str]] = {
            file: file.read_text() if file.exists() else None
            for file in (metadata_yaml, config_yaml, actions_yaml)
        }

        any_metadata_files_present_in_charm_virtual_root = any(
            v is not None for v in metadata_files_present.values()
        )

        if spec.is_autoloaded and charm_virtual_root_is_custom:
            # since the spec is autoloaded, in theory the metadata contents won't differ, so we can
            # overwrite away even if the custom vroot is the real charm root (the local repo).
            # Still, log it for clarity.
            if any_metadata_files_present_in_charm_virtual_root:
                logger.debug(
                    f"metadata files found in custom charm_root {charm_virtual_root}. "
                    f"The spec was autoloaded so the contents should be identical. "
                    f"Proceeding...",
                )

        elif (
            not spec.is_autoloaded and any_metadata_files_present_in_charm_virtual_root
        ):
            logger.warn(
                f"Some metadata files found in custom user-provided charm_root "
                f"{charm_virtual_root} while you have passed meta, config or actions to "
                f"Context.run(). "
                "Single source of truth are the arguments passed to Context.run(). "
                "charm_root metadata files will be overwritten for the "
                "duration of this test, and restored afterwards. "
                "To avoid this, clean any metadata files from the charm_root before calling run.",
            )

        metadata_yaml.write_text(yaml.safe_dump(spec.meta))
        config_yaml.write_text(yaml.safe_dump(spec.config or {}))
        actions_yaml.write_text(yaml.safe_dump(spec.actions or {}))

        yield virtual_charm_root

        if charm_virtual_root_is_custom:
            for file, previous_content in metadata_files_present.items():
                if previous_content is None:  # None == file did not exist before
                    file.unlink()
                else:
                    file.write_text(previous_content)

        else:
            charm_virtual_root.cleanup()

    @staticmethod
    def _get_state_db(temporary_charm_root: Path):
        charm_state_path = temporary_charm_root / ".unit-state.db"
        return UnitStateDB(charm_state_path)

    def _initialize_storage(self, state: "State", temporary_charm_root: Path):
        """Before we start processing this event, store the relevant parts of State."""
        store = self._get_state_db(temporary_charm_root)
        store.apply_state(state)

    def _close_storage(self, state: "State", temporary_charm_root: Path):
        """Now that we're done processing this event, read the charm state and expose it."""
        store = self._get_state_db(temporary_charm_root)
        deferred = store.get_deferred_events()
        stored_state = store.get_stored_state()
        return state.replace(deferred=deferred, stored_state=stored_state)

    @contextmanager
    def _exec_ctx(self, ctx: "Context") -> ContextManager[Tuple[Path, List[EventBase]]]:
        """python 3.8 compatibility shim"""
        with self._virtual_charm_root() as temporary_charm_root:
            # todo allow customizing capture_events
            with capture_events(
                include_deferred=ctx.capture_deferred_events,
                include_framework=ctx.capture_framework_events,
            ) as captured:
                yield (temporary_charm_root, captured)

    @contextmanager
    def exec(
        self,
        state: "State",
        event: "Event",
        context: "Context",
    ) -> ContextManager["Ops"]:
        """Runs an event with this state as initial state on a charm.

        Returns the 'output state', that is, the state as mutated by the charm during the
        event handling.

        This will set the environment up and call ops.main.main().
        After that it's up to ops.
        """
        # todo consider forking out a real subprocess and do the mocking by
        #  mocking hook tool executables

        from scenario.consistency_checker import check_consistency  # avoid cycles

        check_consistency(state, event, self._charm_spec, self._juju_version)

        charm_type = self._charm_spec.charm_type
        logger.info(f"Preparing to fire {event.name} on {charm_type.__name__}")

        # we make a copy to avoid mutating the input state
        output_state = state.copy()

        logger.info(" - generating virtual charm root")
        with self._exec_ctx(context) as (temporary_charm_root, captured):
            logger.info(" - initializing storage")
            self._initialize_storage(state, temporary_charm_root)

            logger.info(" - preparing env")
            env = self._get_event_env(
                state=state,
                event=event,
                charm_root=temporary_charm_root,
            )
            os.environ.update(env)

            logger.info(" - Entering ops.main (mocked).")
            from scenario.ops_main_mock import Ops

            try:
                ops = Ops(
                    state=output_state,
                    event=event,
                    context=context,
                    charm_spec=self._charm_spec.replace(
                        charm_type=self._wrap(charm_type),
                    ),
                )
                ops.setup()

                yield ops

                # if the caller did not manually emit or commit: do that.
                ops.finalize()

            except NoObserverError:
                raise  # propagate along
            except Exception as e:
                raise UncaughtCharmError(
                    f"Uncaught exception ({type(e)}) in operator/charm code: {e!r}",
                ) from e

            finally:
                logger.info(" - Exited ops.main.")

                logger.info(" - Clearing env")
                self._cleanup_env(env)

            logger.info(" - closing storage")
            output_state = self._close_storage(output_state, temporary_charm_root)

        context.emitted_events.extend(captured)

        logger.info("event dispatched. done.")
        context._set_output_state(output_state)



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
class SnapshotError(RuntimeError):
    """Base class for errors raised by snapshot."""


class InvalidTargetUnitName(SnapshotError):
    """Raised if the unit name passed to snapshot is invalid."""


class InvalidTargetModelName(SnapshotError):
    """Raised if the model name passed to snapshot is invalid."""


class StateApplyError(SnapshotError):
    """Raised when the state-apply juju command fails."""



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
from importlib import metadata
from importlib.metadata import PackageNotFoundError
from pathlib import Path

import typer

from scenario.scripts import logger
from scenario.scripts.snapshot import snapshot
from scenario.scripts.state_apply import state_apply


def _version():
    """Print the scenario version and exit."""
    try:
        print(metadata.version("ops-scenario"))
        return
    except PackageNotFoundError:
        pass

    pyproject_toml = Path(__file__).parent.parent.parent / "pyproject.toml"

    if not pyproject_toml.exists():
        print("<weird>")
        return

    for line in pyproject_toml.read_text().split("\n"):
        if line.startswith("version"):
            print(line.split("=")[1].strip("\"' "))
            return


def main():
    app = typer.Typer(
        name="scenario",
        help="Scenario utilities. "
        "For docs, issues and feature requests, visit "
        "the github repo --> https://github.com/canonical/ops-scenario",
        no_args_is_help=True,
        rich_markup_mode="markdown",
    )

    app.command(name="version")(_version)
    app.command(name="snapshot", no_args_is_help=True)(snapshot)
    app.command(name="state-apply", no_args_is_help=True)(state_apply)

    @app.callback()
    def setup_logging(verbose: int = typer.Option(0, "-v", count=True)):
        logger.setup_logging(verbose)

    app()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

import json
import logging
import os
import sys
from pathlib import Path
from subprocess import CalledProcessError, run
from typing import Dict, Iterable, List, Optional

import typer

from scenario.scripts.errors import InvalidTargetUnitName, StateApplyError
from scenario.scripts.utils import JujuUnitName
from scenario.state import (
    Container,
    DeferredEvent,
    Port,
    Relation,
    Secret,
    State,
    StoredState,
    _EntityStatus,
)

SNAPSHOT_DATA_DIR = (Path(os.getcwd()).parent / "snapshot_storage").absolute()

logger = logging.getLogger("snapshot")


def set_relations(relations: Iterable[Relation]) -> List[str]:  # noqa: U100
    logger.info("preparing relations...")
    logger.warning("set_relations not implemented yet")
    return []


def set_status(
    unit_status: _EntityStatus,
    app_status: _EntityStatus,
    app_version: str,
) -> List[str]:
    logger.info("preparing status...")
    cmds = []

    cmds.append(f"status-set {unit_status.name} {unit_status.message}")
    cmds.append(f"status-set --application {app_status.name} {app_status.message}")
    cmds.append(f"application-version-set {app_version}")

    return cmds


def set_config(config: Dict[str, str]) -> List[str]:  # noqa: U100
    logger.info("preparing config...")
    logger.warning("set_config not implemented yet")
    return []


def set_opened_ports(opened_ports: List[Port]) -> List[str]:
    logger.info("preparing opened ports...")
    # fixme: this will only open new ports, it will not close all already-open ports.

    cmds = []

    for port in opened_ports:
        cmds.append(f"open-port {port.port}/{port.protocol}")

    return cmds


def set_containers(containers: Iterable[Container]) -> List[str]:  # noqa: U100
    logger.info("preparing containers...")
    logger.warning("set_containers not implemented yet")
    return []


def set_secrets(secrets: Iterable[Secret]) -> List[str]:  # noqa: U100
    logger.info("preparing secrets...")
    logger.warning("set_secrets not implemented yet")
    return []


def set_deferred_events(
    deferred_events: Iterable[DeferredEvent],  # noqa: U100
) -> List[str]:
    logger.info("preparing deferred_events...")
    logger.warning("set_deferred_events not implemented yet")
    return []


def set_stored_state(stored_state: Iterable[StoredState]) -> List[str]:  # noqa: U100
    logger.info("preparing stored_state...")
    logger.warning("set_stored_state not implemented yet")
    return []


def exec_in_unit(target: JujuUnitName, model: str, cmds: List[str]):
    logger.info("Running juju exec...")

    _model = f" -m {model}" if model else ""
    cmd_fmt = "; ".join(cmds)
    try:
        run(f'juju exec -u {target}{_model} -- "{cmd_fmt}"')
    except CalledProcessError as e:
        raise StateApplyError(
            f"Failed to apply state: process exited with {e.returncode}; "
            f"stdout = {e.stdout}; "
            f"stderr = {e.stderr}.",
        )


def run_commands(cmds: List[str]):
    logger.info("Applying remaining state...")
    for cmd in cmds:
        try:
            run(cmd)
        except CalledProcessError as e:
            # todo: should we log and continue instead?
            raise StateApplyError(
                f"Failed to apply state: process exited with {e.returncode}; "
                f"stdout = {e.stdout}; "
                f"stderr = {e.stderr}.",
            )


def _state_apply(
    target: str,
    state: State,
    model: Optional[str] = None,
    include: str = None,
    include_juju_relation_data=False,  # noqa: U100
    push_files: Dict[str, List[Path]] = None,  # noqa: U100
    snapshot_data_dir: Path = SNAPSHOT_DATA_DIR,  # noqa: U100
):
    """see state_apply's docstring"""
    logger.info("Starting state-apply...")

    try:
        target = JujuUnitName(target)
    except InvalidTargetUnitName:
        logger.critical(
            f"invalid target: {target!r} is not a valid unit name. Should be formatted like so:"
            f"`foo/1`, or `database/0`, or `myapp-foo-bar/42`.",
        )
        sys.exit(1)

    logger.info(f'beginning snapshot of {target} in model {model or "<current>"}...')

    def if_include(key, fn):
        if include is None or key in include:
            return fn()
        return []

    j_exec_cmds: List[str] = []

    j_exec_cmds += if_include(
        "s",
        lambda: set_status(state.unit_status, state.app_status, state.workload_version),
    )
    j_exec_cmds += if_include("p", lambda: set_opened_ports(state.opened_ports))
    j_exec_cmds += if_include("r", lambda: set_relations(state.relations))
    j_exec_cmds += if_include("S", lambda: set_secrets(state.secrets))

    cmds: List[str] = []

    # todo: config is a bit special because it's not owned by the unit but by the cloud admin.
    #  should it be included in state-apply?
    # if_include("c", lambda: set_config(state.config))
    cmds += if_include("k", lambda: set_containers(state.containers))
    cmds += if_include("d", lambda: set_deferred_events(state.deferred))
    cmds += if_include("t", lambda: set_stored_state(state.stored_state))

    # we gather juju-exec commands to run them all at once in the unit.
    exec_in_unit(target, model, j_exec_cmds)
    # non-juju-exec commands are ran one by one, individually
    run_commands(cmds)

    logger.info("Done!")


def state_apply(
    target: str = typer.Argument(..., help="Target unit."),
    state: Path = typer.Argument(
        ...,
        help="Source State to apply. Json file containing a State data structure; "
        "the same you would obtain by running snapshot.",
    ),
    model: Optional[str] = typer.Option(
        None,
        "-m",
        "--model",
        help="Which model to look at.",
    ),
    include: str = typer.Option(
        "scrkSdt",
        "--include",
        "-i",
        help="What parts of the state to apply. Defaults to: all of them. "
        "``r``: relation, ``c``: config, ``k``: containers, "
        "``s``: status, ``S``: secrets(!), "
        "``d``: deferred events, ``t``: stored state.",
    ),
    include_juju_relation_data: bool = typer.Option(
        False,
        "--include-juju-relation-data",
        help="Whether to include in the relation data the default juju keys (egress-subnets,"
        "ingress-address, private-address).",
        is_flag=True,
    ),
    push_files: Path = typer.Option(
        None,
        "--push-files",
        help="Path to a local file containing a json spec of files to be fetched from the unit. "
        "For k8s units, it's supposed to be a {container_name: List[Path]} mapping listing "
        "the files that need to be pushed to the each container.",
    ),
    # TODO: generalize "push_files" to allow passing '.' for the 'charm' container or 'the machine'.
    data_dir: Path = typer.Option(
        SNAPSHOT_DATA_DIR,
        "--data-dir",
        help="Directory in which to any files associated with the state are stored. In the case "
        "of k8s charms, this might mean files obtained through Mounts,",
    ),
):
    """Apply a State to a remote target unit.

    If black is available, the output will be piped through it for formatting.

    Usage: state-apply myapp/0 > ./tests/scenario/case1.py
    """
    push_files_ = json.loads(push_files.read_text()) if push_files else None
    state_json = json.loads(state.read_text())

    # TODO: state_json to State
    raise NotImplementedError("WIP: implement State.from_json")
    state_: State = State.from_json(state_json)

    return _state_apply(
        target=target,
        state=state_,
        model=model,
        include=include,
        include_juju_relation_data=include_juju_relation_data,
        snapshot_data_dir=data_dir,
        push_files=push_files_,
    )


# for the benefit of scripted usage
_state_apply.__doc__ = state_apply.__doc__

if __name__ == "__main__":
    from scenario import State

    _state_apply("zookeeper/0", model="foo", state=State())



import pytest
from ops.charm import CharmBase
from ops.framework import Framework
from ops.model import (
    ActiveStatus,
    BlockedStatus,
    ErrorStatus,
    MaintenanceStatus,
    UnknownStatus,
    WaitingStatus,
)

from scenario import Context
from scenario.state import State, _status_to_entitystatus
from tests.helpers import trigger


@pytest.fixture(scope="function")
def mycharm():
    class MyCharm(CharmBase):
        def __init__(self, framework: Framework):
            super().__init__(framework)
            for evt in self.on.events().values():
                self.framework.observe(evt, self._on_event)

        def _on_event(self, event):
            pass

    return MyCharm


def test_initial_status(mycharm):
    def post_event(charm: CharmBase):
        assert charm.unit.status == UnknownStatus()

    out = trigger(
        State(leader=True),
        "update_status",
        mycharm,
        meta={"name": "local"},
        post_event=post_event,
    )

    assert out.unit_status == UnknownStatus()


def test_status_history(mycharm):
    def post_event(charm: CharmBase):
        for obj in [charm.unit, charm.app]:
            obj.status = ActiveStatus("1")
            obj.status = BlockedStatus("2")
            obj.status = WaitingStatus("3")

    ctx = Context(
        mycharm,
        meta={"name": "local"},
    )

    out = ctx.run(
        "update_status",
        State(leader=True),
        post_event=post_event,
    )

    assert out.unit_status == WaitingStatus("3")
    assert ctx.unit_status_history == [
        UnknownStatus(),
        ActiveStatus("1"),
        BlockedStatus("2"),
    ]

    assert out.app_status == WaitingStatus("3")
    assert ctx.app_status_history == [
        UnknownStatus(),
        ActiveStatus("1"),
        BlockedStatus("2"),
    ]


def test_status_history_preservation(mycharm):
    def post_event(charm: CharmBase):
        for obj in [charm.unit, charm.app]:
            obj.status = WaitingStatus("3")

    ctx = Context(
        mycharm,
        meta={"name": "local"},
    )

    out = ctx.run(
        "update_status",
        State(
            leader=True,
            unit_status=ActiveStatus("foo"),
            app_status=ActiveStatus("bar"),
        ),
        post_event=post_event,
    )

    assert out.unit_status == WaitingStatus("3")
    assert ctx.unit_status_history == [ActiveStatus("foo")]

    assert out.app_status == WaitingStatus("3")
    assert ctx.app_status_history == [ActiveStatus("bar")]


def test_workload_history(mycharm):
    def post_event(charm: CharmBase):
        charm.unit.set_workload_version("1")
        charm.unit.set_workload_version("1.1")
        charm.unit.set_workload_version("1.2")

    ctx = Context(
        mycharm,
        meta={"name": "local"},
    )

    out = ctx.run(
        "update_status",
        State(
            leader=True,
        ),
        post_event=post_event,
    )

    assert ctx.workload_version_history == ["1", "1.1"]
    assert out.workload_version == "1.2"


@pytest.mark.parametrize(
    "status",
    (
        ActiveStatus("foo"),
        WaitingStatus("bar"),
        BlockedStatus("baz"),
        MaintenanceStatus("qux"),
        ErrorStatus("fiz"),
        UnknownStatus(),
    ),
)
def test_status_comparison(status):
    entitystatus = _status_to_entitystatus(status)
    assert entitystatus == entitystatus == status
    assert isinstance(entitystatus, type(status))
    assert repr(entitystatus) == repr(status)



import importlib
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Type

import pytest
import yaml
from ops.testing import CharmType

from scenario import Context, Relation, State
from scenario.context import ContextSetupError

CHARM = """
from ops import CharmBase

class MyCharm(CharmBase): pass
"""


@contextmanager
def import_name(name: str, source: Path) -> Type[CharmType]:
    pkg_path = str(source.parent)
    sys.path.append(pkg_path)
    charm = importlib.import_module("charm")
    obj = getattr(charm, name)
    sys.path.remove(pkg_path)
    yield obj
    del sys.modules["charm"]


@contextmanager
def create_tempcharm(
    root: Path,
    charm: str = CHARM,
    meta=None,
    actions=None,
    config=None,
    name: str = "MyCharm",
):
    src = root / "src"
    src.mkdir(parents=True)
    charmpy = src / "charm.py"
    charmpy.write_text(charm)

    if meta is not None:
        (root / "metadata.yaml").write_text(yaml.safe_dump(meta))

    if actions is not None:
        (root / "actions.yaml").write_text(yaml.safe_dump(actions))

    if config is not None:
        (root / "config.yaml").write_text(yaml.safe_dump(config))

    with import_name(name, charmpy) as charm:
        yield charm


def test_meta_autoload(tmp_path):
    with create_tempcharm(tmp_path, meta={"name": "foo"}) as charm:
        ctx = Context(charm)
        ctx.run("start", State())


def test_no_meta_raises(tmp_path):
    with create_tempcharm(
        tmp_path,
    ) as charm:
        # metadata not found:
        with pytest.raises(ContextSetupError):
            Context(charm)


def test_relations_ok(tmp_path):
    with create_tempcharm(
        tmp_path, meta={"name": "josh", "requires": {"cuddles": {"interface": "arms"}}}
    ) as charm:
        # this would fail if there were no 'cuddles' relation defined in meta
        Context(charm).run("start", State(relations=[Relation("cuddles")]))


def test_config_defaults(tmp_path):
    with create_tempcharm(
        tmp_path,
        meta={"name": "josh"},
        config={"options": {"foo": {"type": "bool", "default": True}}},
    ) as charm:
        # this would fail if there were no 'cuddles' relation defined in meta
        with Context(charm).manager("start", State()) as mgr:
            mgr.run()
            assert mgr.charm.config["foo"] is True



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.
from scenario.context import ActionOutput, Context
from scenario.state import (
    Action,
    Address,
    BindAddress,
    Container,
    DeferredEvent,
    Event,
    ExecOutput,
    InjectRelation,
    Model,
    Mount,
    Network,
    PeerRelation,
    Port,
    Relation,
    RelationBase,
    Secret,
    State,
    StateValidationError,
    Storage,
    StoredState,
    SubordinateRelation,
    deferred,
)

__all__ = [
    "Action",
    "ActionOutput",
    "Context",
    "deferred",
    "StateValidationError",
    "Secret",
    "RelationBase",
    "Relation",
    "SubordinateRelation",
    "PeerRelation",
    "Model",
    "ExecOutput",
    "Mount",
    "Container",
    "Address",
    "BindAddress",
    "Network",
    "Port",
    "Storage",
    "StoredState",
    "State",
    "DeferredEvent",
    "Event",
    "InjectRelation",
]
